# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Community actions tests."""

import pytest

from invenio_rdm_migrator.load.ids import generate_pk_for, generate_uuid
from invenio_rdm_migrator.load.postgresql.transactions.operations import OperationType
from invenio_rdm_migrator.streams.actions.load import (
    CommunityCreateAction,
    CommunityDeleteAction,
    CommunityUpdateAction,
)
from invenio_rdm_migrator.streams.models.communities import (
    Community,
    CommunityFile,
    CommunityMember,
)
from invenio_rdm_migrator.streams.models.files import (
    FilesBucket,
    FilesInstance,
    FilesObjectVersion,
)
from invenio_rdm_migrator.streams.models.oai import OAISet


@pytest.fixture(scope="function")
def community_data():
    """Community data."""
    return {
        "slug": "test-community",
        "created": "2021-05-01T00:00:00",
        "updated": "2021-05-01T00:00:00",
        "json": {
            "files": {"enabled": True},
            "access": {
                "visibility": "public",
                "member_policy": "open",
                "record_policy": "open",
            },
            "metadata": {
                "page": "<p>This is the community page</p>",
                "title": "Test community",
                "description": "Test community description",
                "curation_policy": "<p>Curation policy for test community</p>",
            },
        },
        "version_id": 1,
    }


@pytest.fixture()
def owner_data():
    """Community owner data."""
    return {
        "created": "2021-05-01T00:00:00",
        "updated": "2021-05-01T00:00:00",
        "json": {},
        "version_id": 1,
        "role": "owner",
        "visible": True,
        "active": True,
        "user_id": 1234,
        "group_id": None,
        "request_id": None,
    }


@pytest.fixture(scope="function")
def oai_set_data():
    """OAI set data."""
    return {
        "created": "2023-06-29T13:00:00",
        "updated": "2023-06-29T14:00:00",
        "spec": "community-test-community",
        "name": "Test community",
        "description": "Test community description",
    }


@pytest.fixture(scope="function")
def bucket_data():
    """Bucket data."""
    return {
        "created": "2023-06-29T13:00:00",
        "updated": "2023-06-29T14:00:00",
        "default_location": 1,
        "default_storage_class": "S",
        "size": 1562554,
        "quota_size": 50000000000,
        "max_file_size": 50000000000,
        "locked": False,
        "deleted": False,
    }


@pytest.fixture(scope="function")
def fi_data():
    """File instance data."""
    return {
        "id": "e94b243e-9c0c-44df-bd1f-6decc374cf78",
        "created": "2023-06-29T13:00:00",
        "updated": "2023-06-29T14:00:00",
        "uri": "root://eosmedia.cern.ch//eos/media/zenodo/test/data/e9/4b/243e-9c0c-44df-bd1f-6decc374cf78/data",
        "storage_class": "S",
        "size": 1562554,
        "checksum": "md5:3cc016be06f2be46d3a438db23c40bf3",
        "readable": True,
        "writable": False,
        "last_check_at": None,
        "last_check": True,
    }


@pytest.fixture(scope="function")
def ov_data(fi_data):
    """Object version data."""
    return {
        "version_id": "f8200dc7-55b6-4785-abd0-f3d13b143c98",
        "created": "2023-06-29T13:00:00",
        "updated": "2023-06-29T14:00:00",
        "key": "logo",
        "bucket_id": "<other_bucket>",
        "file_id": fi_data["id"],
        "_mimetype": None,
        "is_head": True,
    }


@pytest.fixture(scope="function")
def community_file_data():
    """Community file data."""
    return {
        "json": {},
        "record_id": None,
        "version_id": 1,
        "key": "logo",
    }


def test_create_community(
    communities_state,
    community_data,
    owner_data,
    oai_set_data,
    bucket_data,
    fi_data,
    ov_data,
    community_file_data,
):
    """Test community create action."""
    data = dict(
        tx_id=1,
        community=community_data,
        owner=owner_data,
        bucket=bucket_data,
        oai_set=oai_set_data,
        community_file=community_file_data,
        file_instance=fi_data,
        object_version=ov_data,
    )

    action = CommunityCreateAction(data)
    rows = list(action.prepare())
    assert len(rows) == 7

    bucket_row, community_row, oai_set_row, owner_row, fi_row, ov_row, cf_row = rows

    bucket_id = bucket_row.data["id"]
    assert bucket_row.type == OperationType.INSERT
    assert bucket_row.model == FilesBucket

    c_id = community_row.data["id"]
    slug = community_row.data["slug"]
    assert community_row.type == OperationType.INSERT
    assert community_row.model == Community
    assert community_row.data["bucket_id"] == bucket_id

    oai_set_id = oai_set_row.data["id"]
    assert oai_set_row.type == OperationType.INSERT
    assert oai_set_row.model == OAISet
    assert oai_set_id
    assert oai_set_row.data["search_pattern"] == f"parent.communities.ids:{c_id}"
    assert oai_set_row.data["system_created"] is True

    assert owner_row.type == OperationType.INSERT
    assert owner_row.model == CommunityMember
    assert owner_row.data["community_id"] == c_id

    assert fi_row.type == OperationType.INSERT
    assert fi_row.model == FilesInstance

    ov_id = ov_row.data["version_id"]
    assert ov_row.type == OperationType.INSERT
    assert ov_row.model == FilesObjectVersion
    assert ov_row.data["bucket_id"] == bucket_id

    cf_id = cf_row.data["id"]
    assert cf_row.type == OperationType.INSERT
    assert cf_row.model == CommunityFile
    assert cf_row.data["record_id"] == c_id
    assert cf_row.data["object_version_id"] == ov_id

    # Community is in the state
    assert communities_state.get(slug) == {
        "slug": slug,
        "id": c_id,
        "bucket_id": bucket_id,
        "oai_set_id": oai_set_id,
        "community_file_id": cf_id,
        "logo_object_version_id": ov_id,
    }


def test_create_community_no_logo(
    communities_state,
    community_data,
    owner_data,
    oai_set_data,
    bucket_data,
):
    """Test community create action w/o logo."""
    data = dict(
        tx_id=1,
        community=community_data,
        owner=owner_data,
        bucket=bucket_data,
        oai_set=oai_set_data,
    )

    action = CommunityCreateAction(data)
    rows = list(action.prepare())
    assert len(rows) == 4

    bucket_row, community_row, oai_set_row, owner_row = rows

    bucket_id = bucket_row.data["id"]
    assert bucket_row.type == OperationType.INSERT
    assert bucket_row.model == FilesBucket

    c_id = community_row.data["id"]
    assert community_row.type == OperationType.INSERT
    assert community_row.model == Community
    assert community_row.data["bucket_id"] == bucket_id

    assert oai_set_row.type == OperationType.INSERT
    assert oai_set_row.model == OAISet
    assert oai_set_row.data["id"]
    assert oai_set_row.data["search_pattern"] == f"parent.communities.ids:{c_id}"
    assert oai_set_row.data["system_created"] is True

    assert owner_row.type == OperationType.INSERT
    assert owner_row.model == CommunityMember
    assert owner_row.data["community_id"] == c_id

    # Community is in the state
    assert communities_state.get(community_row.data["slug"]) == {
        "slug": community_row.data["slug"],
        "id": c_id,
        "bucket_id": bucket_id,
        "oai_set_id": oai_set_row.data["id"],
        "community_file_id": None,
        "logo_object_version_id": None,
    }


def test_update_community(
    communities_state,
    community_data,
    fi_data,
    ov_data,
    community_file_data,
):
    """Test community update action."""
    # Use a community that's already in the state
    slug = community_data["slug"] = "comm"
    state = communities_state.get(slug)
    community_id = state["id"]
    bucket_id = state["bucket_id"]
    oai_set_id = state["oai_set_id"]

    # remove/modify some keys from the community
    community_data.pop("created")
    community_data.pop("version_id")
    community_data["json"]["metadata"]["title"] = "Test community (v2)"

    data = dict(
        tx_id=1,
        community=community_data,
        community_file=community_file_data,
        file_instance=fi_data,
        object_version=ov_data,
    )
    action = CommunityUpdateAction(data)
    rows = list(action.prepare())
    assert len(rows) == 4

    community_row, fi_row, ov_row, cf_row = rows

    assert community_row.type == OperationType.UPDATE
    assert community_row.model == Community
    assert community_row.data["id"] == community_id
    # Only the PK and the updated values ae present
    assert community_row.data.keys() == {"id", "updated", "json", "slug"}

    assert fi_row.type == OperationType.INSERT
    assert fi_row.model == FilesInstance

    ov_id = ov_row.data["version_id"]
    assert ov_row.type == OperationType.INSERT
    assert ov_row.model == FilesObjectVersion
    assert ov_row.data["bucket_id"] == bucket_id

    cf_id = cf_row.data["id"]
    assert cf_row.type == OperationType.INSERT
    assert cf_row.model == CommunityFile
    assert cf_row.data["record_id"] == community_id
    assert cf_row.data["object_version_id"] == ov_id

    # Community state has been updated with the file PKs
    assert communities_state.get(slug) == {
        "slug": slug,
        "id": community_id,
        "bucket_id": bucket_id,
        "oai_set_id": oai_set_id,
        "community_file_id": cf_id,
        "logo_object_version_id": ov_id,
    }


def test_update_community_only_metadata(
    communities_state,
    community_data,
):
    """Test community update action for only metadata."""
    # Use a community that's already in the state
    slug = community_data["slug"] = "comm"
    state = communities_state.get(slug)
    community_id = state["id"]
    bucket_id = state["bucket_id"]
    oai_set_id = state["oai_set_id"]

    # remove/modify some keys from the community
    community_data.pop("created")
    community_data.pop("version_id")
    community_data["json"]["metadata"]["title"] = "Test community (v2)"

    data = dict(tx_id=1, community=community_data)
    action = CommunityUpdateAction(data)
    rows = list(action.prepare())
    assert len(rows) == 1

    (community_row,) = rows

    assert community_row.type == OperationType.UPDATE
    assert community_row.model == Community
    assert community_row.data["id"] == community_id
    assert community_row.data.keys() == {"id", "updated", "json", "slug"}

    # Community state has been updated with the file PKs
    assert communities_state.get(slug) == {
        "slug": slug,
        "id": community_id,
        "bucket_id": bucket_id,
        "oai_set_id": oai_set_id,
        "community_file_id": None,
        "logo_object_version_id": None,
    }


def test_update_community_update_existing_logo(
    communities_state,
    community_data,
    fi_data,
    ov_data,
    community_file_data,
):
    """Test community update action for a community with an existing logo."""
    # Use a community that's already in the state
    slug = community_data["slug"] = "comm"
    state = communities_state.get(slug)
    community_id = state["id"]
    bucket_id = state["bucket_id"]
    oai_set_id = state["oai_set_id"]

    # Add existing logo file PKs
    community_file_id = generate_uuid(None)
    old_logo_object_version_id = generate_uuid(None)
    communities_state.update(
        slug,
        {
            "community_file_id": community_file_id,
            "logo_object_version_id": old_logo_object_version_id,
        },
    )

    data = dict(
        tx_id=1,
        community=community_data,
        community_file=community_file_data,
        file_instance=fi_data,
        object_version=ov_data,
    )
    action = CommunityUpdateAction(data)
    rows = list(action.prepare())
    assert len(rows) == 5

    community_row, fi_row, old_ov_row, new_ov_row, cf_row = rows

    assert community_row.type == OperationType.UPDATE
    assert community_row.model == Community
    assert community_row.data["id"] == community_id

    assert fi_row.type == OperationType.INSERT
    assert fi_row.model == FilesInstance

    old_ov_row.data["version_id"] == old_logo_object_version_id
    assert old_ov_row.type == OperationType.UPDATE
    assert old_ov_row.model == FilesObjectVersion
    assert old_ov_row.data["is_head"] is False

    new_ov_id = new_ov_row.data["version_id"]
    assert new_ov_id != old_logo_object_version_id
    assert new_ov_row.type == OperationType.INSERT
    assert new_ov_row.model == FilesObjectVersion
    assert new_ov_row.data["bucket_id"] == bucket_id

    assert cf_row.data["id"] == community_file_id
    assert cf_row.type == OperationType.UPDATE
    assert cf_row.model == CommunityFile
    assert cf_row.data["record_id"] == community_id
    assert cf_row.data["object_version_id"] == new_ov_id

    # Community state has been updated with the file PKs
    assert communities_state.get(slug) == {
        "slug": slug,
        "id": community_id,
        "bucket_id": bucket_id,
        "oai_set_id": oai_set_id,
        "community_file_id": community_file_id,
        "logo_object_version_id": new_ov_id,
    }


def test_delete_community(
    communities_state,
    community_data,
):
    """Test community delete action."""
    # Use a community that's already in the state
    slug = community_data["slug"] = "comm"
    state = communities_state.get(slug)
    community_id = state["id"]
    oai_set_id = state["oai_set_id"]

    data = dict(tx_id=1, community=community_data)
    action = CommunityDeleteAction(data)
    rows = list(action.prepare())
    assert len(rows) == 2

    community_row, oai_set_row = rows

    assert community_row.type == OperationType.UPDATE
    assert community_row.model == Community
    assert community_row.data["id"] == community_id
    assert community_row.data["json"] is None

    assert oai_set_row.type == OperationType.DELETE
    assert oai_set_row.model == OAISet
    assert oai_set_row.data["id"] == oai_set_id

    # Community state is not deleted
    assert communities_state.get(slug)


def test_delete_community_with_logo(
    communities_state,
    community_data,
):
    """Test community delete action for community with logo."""
    # Use a community that's already in the state
    slug = community_data["slug"] = "comm"
    state = communities_state.get(slug)
    community_id = state["id"]
    oai_set_id = state["oai_set_id"]

    # Add existing logo file PKs
    community_file_id = generate_uuid(None)
    logo_object_version_id = generate_uuid(None)
    communities_state.update(
        slug,
        {
            "community_file_id": community_file_id,
            "logo_object_version_id": logo_object_version_id,
        },
    )

    data = dict(tx_id=1, community=community_data)
    action = CommunityDeleteAction(data)
    rows = list(action.prepare())
    assert len(rows) == 4

    community_row, oai_set_row, ov_row, cf_row = rows

    assert community_row.type == OperationType.UPDATE
    assert community_row.model == Community
    assert community_row.data["id"] == community_id
    assert community_row.data["json"] is None

    assert oai_set_row.type == OperationType.DELETE
    assert oai_set_row.model == OAISet
    assert oai_set_row.data["id"] == oai_set_id

    assert ov_row.type == OperationType.DELETE
    assert ov_row.model == FilesObjectVersion
    assert ov_row.data["version_id"] == logo_object_version_id

    assert cf_row.type == OperationType.DELETE
    assert cf_row.model == CommunityFile
    assert cf_row.data["id"] == community_file_id

    # Community state is not deleted
    assert communities_state.get(slug)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Communities actions module."""

from dataclasses import dataclass
from typing import Optional

from ....actions import LoadAction, LoadData
from ....load.ids import generate_pk_for, generate_uuid
from ....load.postgresql.transactions.operations import Operation, OperationType
from ....state import STATE
from ...models.communities import Community, CommunityFile, CommunityMember
from ...models.files import FilesBucket, FilesInstance, FilesObjectVersion
from ...models.oai import OAISet


@dataclass
class CommunityData(LoadData):
    """Community action data."""

    community: dict
    owner: dict
    oai_set: dict
    bucket: dict

    community_file: Optional[dict] = None
    file_instance: Optional[dict] = None
    object_version: Optional[dict] = None


class CommunityCreateAction(LoadAction):
    """Create a community."""

    pks = [
        ("community", "id", generate_uuid),
        ("owner", "id", generate_uuid),
        ("oai_set", "id", generate_pk_for(OAISet)),
        ("bucket", "id", generate_uuid),
        ("community_file", "id", generate_uuid),
    ]

    name = "community-create"
    data_cls = CommunityData

    def _generate_rows(self, **kwargs):
        """Generates rows for a new community."""
        community_id = self.data.community["id"]
        slug = self.data.community["slug"]
        community_file_id = None
        logo_object_version_id = None

        # Fill-in the community bucket
        bucket_id = self.data.bucket["id"]
        self.data.community["bucket_id"] = bucket_id

        yield Operation(OperationType.INSERT, FilesBucket, self.data.bucket)
        yield Operation(OperationType.INSERT, Community, self.data.community)

        oai_set_id = self.data.oai_set["id"]
        self.data.oai_set["search_pattern"] = f"parent.communities.ids:{community_id}"
        self.data.oai_set["system_created"] = True
        yield Operation(OperationType.INSERT, OAISet, self.data.oai_set)

        # Generate owner member
        self.data.owner["community_id"] = community_id
        yield Operation(OperationType.INSERT, CommunityMember, self.data.owner)

        # Not every community has a logo
        if self.data.community_file:
            yield Operation(
                OperationType.INSERT,
                FilesInstance,
                self.data.file_instance,
            )

            object_version = self.data.object_version
            logo_object_version_id = object_version["version_id"]
            object_version["bucket_id"] = bucket_id
            yield Operation(OperationType.INSERT, FilesObjectVersion, object_version)

            community_file = self.data.community_file
            community_file_id = community_file["id"]
            community_file["record_id"] = community_id
            community_file["object_version_id"] = logo_object_version_id
            yield Operation(OperationType.INSERT, CommunityFile, community_file)

        # Add to state
        STATE.COMMUNITIES.add(
            slug,
            {
                "id": community_id,
                "bucket_id": bucket_id,
                "oai_set_id": oai_set_id,
                "community_file_id": community_file_id,
                "logo_object_version_id": logo_object_version_id,
            },
        )


@dataclass
class CommunityUpdateData(LoadData):
    """Community action data."""

    community: dict

    community_file: Optional[dict] = None
    file_instance: Optional[dict] = None
    object_version: Optional[dict] = None

    # Filled-in by the action
    old_object_version_id: Optional[str] = None


class CommunityUpdateAction(LoadAction):
    """Update a community."""

    name = "community-update"
    data_cls = CommunityUpdateData

    pks = [
        # We generate a ComunityFile.id, just in case there was no logo before
        ("community_file", "id", generate_uuid),
    ]

    def _resolve_references(self, **kwargs):
        slug = self.data.community["slug"]
        state = STATE.COMMUNITIES.get(slug)
        self.data.community.setdefault("id", state["id"])

        # if it's also a logo update we need to resolve more PKs/FKs
        if self.data.community_file:
            self.data.old_object_version_id = state["logo_object_version_id"]
            self.data.object_version["bucket_id"] = state["bucket_id"]

            # If there was a logo already, we reuse the community file PK
            if state["community_file_id"]:
                self.data.community_file["id"] = state["community_file_id"]
            self.data.community_file["record_id"] = state["id"]

    def _generate_rows(self, **kwargs):
        """Generates rows for a new community."""
        yield Operation(OperationType.UPDATE, Community, self.data.community)

        # Not every community update has a logo
        if self.data.community_file:
            yield Operation(
                OperationType.INSERT,
                FilesInstance,
                self.data.file_instance,
            )

            # update the old object version also if there was one
            if self.data.old_object_version_id:
                yield Operation(
                    OperationType.UPDATE,
                    FilesObjectVersion,
                    {"version_id": self.data.old_object_version_id, "is_head": False},
                )

            yield Operation(
                OperationType.INSERT,
                FilesObjectVersion,
                self.data.object_version,
            )

            self.data.community_file["object_version_id"] = self.data.object_version[
                "version_id"
            ]
            if self.data.old_object_version_id:
                # If there was already a logo we UPDATE the community file
                yield Operation(
                    OperationType.UPDATE,
                    CommunityFile,
                    self.data.community_file,
                )
            else:
                # If no logo before we INSERT a community file
                yield Operation(
                    OperationType.INSERT,
                    CommunityFile,
                    self.data.community_file,
                )

            # Update the logo state
            STATE.COMMUNITIES.update(
                self.data.community["slug"],
                {
                    "community_file_id": self.data.community_file["id"],
                    "logo_object_version_id": self.data.object_version["version_id"],
                },
            )


@dataclass
class CommunityDeleteData(LoadData):
    """Community action data."""

    community: dict

    # Filled-in by the action
    oai_set_id: Optional[int] = None
    community_file_id: Optional[str] = None
    object_version_id: Optional[str] = None


class CommunityDeleteAction(LoadAction):
    """Delete a community."""

    name = "community-delete"
    data_cls = CommunityDeleteData

    def _resolve_references(self, **kwargs):
        slug = self.data.community["slug"]
        state = STATE.COMMUNITIES.get(slug)
        self.data.community.setdefault("id", state["id"])
        self.data.oai_set_id = state["oai_set_id"]
        self.data.community_file_id = state["community_file_id"]
        self.data.object_version_id = state["logo_object_version_id"]

    def _generate_rows(self, **kwargs):
        """Generates rows for deleting a communitoy."""
        # Only soft-delete by setting `json = NULL`
        self.data.community["json"] = None
        yield Operation(OperationType.UPDATE, Community, self.data.community)

        # Hard-delete the OAISet
        yield Operation(OperationType.DELETE, OAISet, {"id": self.data.oai_set_id})

        # Clean-up the logo also if there is one
        if self.data.community_file_id:
            yield Operation(
                OperationType.DELETE,
                FilesObjectVersion,
                {"version_id": self.data.object_version_id},
            )
            yield Operation(
                OperationType.DELETE,
                CommunityFile,
                {"id": self.data.community_file_id},
            )

        # NOTE: We don't update the state actually


class CommunityRecordAcceptAction(LoadAction):
    """Accept a record into a community."""

    # TODO: Implement
    pass


class CommunityRecordRejectAction(LoadAction):
    """Reject a record from a community."""

    # TODO: Implement
    pass


class CommunityRecordRemoveAction(LoadAction):
    """Remove a record from a community."""

    # TODO: Implement
    pass



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration GitHub stream."""

from .load import (
    ExistingGitHubRepositoriesCopyLoad,
    ExistingWebhookEventsCopyLoad,
    GitHubReleasesCopyLoad,
)
from .transform import GitHubReleaseTransform

__all__ = (
    "ExistingWebhookEventsCopyLoad",
    "ExistingGitHubRepositoriesCopyLoad",
    "GitHubReleasesCopyLoad",
    "GitHubReleaseTransform",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration users stream."""

from .load import ExistingFilesLoad, FilesCopyLoad
from .transform import (
    FilesBucketEntry,
    FilesInstanceEntry,
    FilesObjectVersionEntry,
    FilesTransform,
)

__all__ = (
    "ExistingFilesLoad",
    "FilesCopyLoad",
    "FilesBucketEntry",
    "FilesInstanceEntry",
    "FilesObjectVersionEntry",
    "FilesTransform",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration record transform interfaces."""

from ...transform import EncryptMixin, IdentityDictKeyMixin, Transform


class OAuthServerTokenTransform(EncryptMixin, Transform, IdentityDictKeyMixin):
    """OAuth server token data transformation."""

    SCOPES_MAPPING = {
        "deposit:actions": None,
        "deposit:extra_formats": None,
        "deposit:write": None,
        "generate": "tokens:generate",
        "user:email": "user:email",
        "webhooks:event": "webhooks:event",
    }
    """Keys are Invenio v3 scopes and values are new RDM scopes."""

    def _access_token(self, entry):
        """Server access token."""
        # as per DB constraints the token can be null even if it does not make sense to be
        token = entry["access_token"]
        return self.re_encrypt(token) if token else None

    def _refresh_token(self, entry):
        """Server refresh token."""
        token = entry["refresh_token"]
        return self.re_encrypt(token) if token else None

    def _scopes(self, entry):
        """Token scopes."""
        scopes = entry.get("_scopes", "").split(" ")
        new_scopes = []
        for scope in scopes:
            new_scope = self.SCOPES_MAPPING[scope]
            if new_scope:
                new_scopes.append(new_scope)

        return " ".join(new_scopes)

    def _transform(self, entry):
        """Transform a single entry."""
        return {
            "id": self._id(entry),
            "client_id": self._client_id(entry),
            "user_id": self._user_id(entry),
            "token_type": self._token_type(entry),
            "access_token": self._access_token(entry),
            "refresh_token": self._refresh_token(entry),
            "expires": self._expires(entry),
            "_scopes": self._scopes(entry),
            "is_personal": self._is_personal(entry),
            "is_internal": self._is_internal(entry),
        }


class OAuthRemoteTokenTransform(Transform, EncryptMixin, IdentityDictKeyMixin):
    """OAuth client remote token data transformation."""

    def _access_token(self, entry):
        """Remote access token."""
        token = entry["access_token"]
        return self.re_encrypt(token)

    def _transform(self, entry):
        """Transform a single entry."""
        return {
            "id_remote_account": self._id_remote_account(entry),
            "token_type": self._token_type(entry),
            "access_token": self._access_token(entry),
            "secret": self._secret(entry),
            "created": self._created(entry),
            "updated": self._updated(entry),
        }



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration record load module."""

from ...load.postgresql.bulk import PostgreSQLCopyLoad
from .table_generators import (
    RDMDraftTableGenerator,
    RDMRecordTableGenerator,
    RDMVersionStateTableGenerator,
)


class RDMRecordCopyLoad(PostgreSQLCopyLoad):
    """PostgreSQL COPY load."""

    def __init__(self, versioning=True, **kwargs):
        """Constructor."""
        table_generators = [RDMRecordTableGenerator(), RDMDraftTableGenerator()]

        if versioning:
            table_generators.append(RDMVersionStateTableGenerator())

        super().__init__(table_generators=table_generators, **kwargs)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Pytest requests configuration."""

from copy import deepcopy

import pytest


@pytest.fixture(scope="module")
def transformed_incl_req_entry():
    """Request data as after being transformed.

    `topic`, `receiver` and `created_by` could be different.
    This is a sample inclusion request for a record into a community.
    """
    return {
        "created": "01/01/2023",
        "updated": "01/01/2023",
        "version_id": "1",
        "json": {
            "type": "community-inclusion",
            "title": "title",
            "topic": {"record": "123456"},
            "status": "submitted",
            "receiver": {"community": "comm"},
            "created_by": {"user": "3"},
            "$schema": "local://requests/request-v1.0.0.json",
        },
        "number": "1",
        "expires_at": "01/01/2023",
    }


@pytest.fixture(scope="module")
def transformed_incl_req_entry_pks(transformed_incl_req_entry):
    """Request data as after being transformed and passed by generate_pks.

    The last step would happen in the table_generator._prepare.
    """
    data = deepcopy(transformed_incl_req_entry)
    data["id"] = "12345678-abcd-1a2b-3c4d-123abc456def"

    return data



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration GtiHub transform interfaces."""

from ...transform import Transform


class GitHubReleaseTransform(Transform):
    """GitHub release transformation."""

    def _created(self, entry):
        """Return created."""
        return entry["created"]

    def _updated(self, entry):
        """Return updated."""
        return entry["updated"]

    def _id(self, entry):
        """Return id."""
        return entry["id"]

    def _release_id(self, entry):
        """Return release_id."""
        return entry["release_id"]

    def _tag(self, entry):
        """Return tag."""
        return entry["tag"]

    def _errors(self, entry):
        """Return errors."""
        return entry["errors"]

    def _repository_id(self, entry):
        """Return repository_id."""
        return entry["repository_id"]

    def _event_id(self, entry):
        """Return event_id."""
        return entry["event_id"]

    def _record_id(self, entry):
        """Return record_id."""
        return None

    def _status(self, entry):
        """Return status."""
        return entry["status"]

    def _transform(self, entry):
        """Transform a single entry."""
        return {
            "created": self._created(entry),
            "updated": self._updated(entry),
            "id": self._id(entry),
            "release_id": self._release_id(entry),
            "tag": self._tag(entry),
            "errors": self._errors(entry),
            "repository_id": self._repository_id(entry),
            "event_id": self._event_id(entry),
            "record_id": self._record_id(entry),
            "recid": entry.get("recid"),
            "status": self._status(entry),
        }



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""PostgreSQL Transaction load operations."""

from enum import Enum

from ..models import Model


class OperationType(str, Enum):
    """SQL operation type enumeration."""

    INSERT = "C"
    UPDATE = "U"
    DELETE = "D"

    def __eq__(self, other):
        """Equality test."""
        return self.value == other.upper()

    def __str__(self):
        """Return its value."""
        return self.value


class Operation:
    """SQL operation.

    A transaction is composed by one or more operations.
    """

    def __init__(self, type: OperationType, model: Model, data: dict):
        """Constructor."""
        self.type = type
        self.model = model
        self.data = data



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""User actions tests."""

import pytest

from invenio_rdm_migrator.load.postgresql.transactions.operations import OperationType
from invenio_rdm_migrator.streams.actions.load import (
    UserDeactivationAction,
    UserEditAction,
    UserRegistrationAction,
)
from invenio_rdm_migrator.streams.models.users import (
    LoginInformation,
    SessionActivity,
    User,
)


@pytest.fixture(scope="function")
def user_data():
    return {
        "id": 123456,
        "created": "2021-05-01T00:00:00",
        "updated": "2021-05-01T00:00:00",
        "username": "test_user",
        "displayname": "test_user",
        "email": "someaddr@domain.org",
        "password": "zmkNzdnG1PXP5C3dmZqlJw==",
        "active": True,
        "confirmed_at": None,
        "version_id": 1,
        "profile": {"full_name": "User test"},
        "preferences": {
            "visibility": "restricted",
            "email_visibility": "restricted",
        },
    }


@pytest.fixture()
def login_info_data():
    return {
        "last_login_at": None,
        "current_login_at": None,
        "last_login_ip": None,
        "current_login_ip": None,
        "login_count": None,
    }


@pytest.fixture()
def sessions_data():
    return [
        {
            "user_id": None,
            "created": 0,
            "updated": 0,
            "sid_s": "bc51d8ea3ccc285c_64cb64fa",
            "ip": None,
            "country": None,
            "browser": None,
            "browser_version": None,
            "os": None,
            "device": None,
        },
        {
            "user_id": None,
            "created": 0,
            "updated": 0,
            "sid_s": "754493997337aa0a_64cb65bc",
            "ip": None,
            "country": None,
            "browser": None,
            "browser_version": None,
            "os": None,
            "device": None,
        },
    ]


def test_register_new_user(secret_keys_state, user_data, login_info_data):
    data = dict(tx_id=1, user=user_data, login_information=login_info_data)
    action = UserRegistrationAction(data)
    rows = list(action.prepare())
    assert len(rows) == 2
    assert rows[0].type == OperationType.INSERT
    assert rows[0].model == User
    assert rows[1].type == OperationType.INSERT
    assert rows[1].model == LoginInformation


def test_edit_user(secret_keys_state, user_data, login_info_data):
    data = dict(tx_id=1, user=user_data, login_information=login_info_data)
    action = UserEditAction(data)
    rows = list(action.prepare())
    assert len(rows) == 2
    assert rows[0].type == OperationType.UPDATE
    assert rows[0].model == User
    assert rows[1].type == OperationType.UPDATE
    assert rows[1].model == LoginInformation


def test_deactivate_user(secret_keys_state, user_data, sessions_data):
    # prepare
    user_data["active"] = False

    # test
    data = dict(tx_id=1, user=user_data, sessions=sessions_data)
    action = UserDeactivationAction(data)
    rows = list(action.prepare())
    assert len(rows) == 3
    assert rows[0].type == OperationType.UPDATE
    assert rows[0].model == User
    assert rows[1].type == OperationType.DELETE
    assert rows[1].model == SessionActivity
    assert rows[2].type == OperationType.DELETE
    assert rows[2].model == SessionActivity



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Pytest configuration."""

import tempfile

import pytest

from invenio_rdm_migrator.state import STATE, StateDB
from invenio_rdm_migrator.streams.records.state import ParentModelValidator

pytest_plugins = ("invenio_rdm_migrator.testutils",)


@pytest.fixture(scope="function")
def tmp_dir():
    """Yields a temporary directory."""
    tmp_dir = tempfile.TemporaryDirectory()
    yield tmp_dir
    tmp_dir.cleanup()


@pytest.fixture(scope="function")
def state(tmp_dir):
    """Yields a state.

    Do not call `save` on this fixture. The in memory database will be reset on each
    function, therefore no information will be persisted from test to test.
    """
    state_db = StateDB(
        db_dir=tmp_dir.name, validators={"parents": ParentModelValidator}
    )
    STATE.initialized_state(state_db, cache=False)

    return STATE


@pytest.fixture(scope="function")
def parents_state(state):
    """Records parent state.

    Keys are concept recids and values are dictionaries.
    """
    state.PARENTS.add(
        "123456",
        {
            "id": "1234abcd-1234-5678-abcd-123abc456def",
            "latest_id": "12345678-abcd-1a2b-3c4d-123abc456def",
            "latest_index": 1,
        },
    )

    return state


@pytest.fixture(scope="function")
def communities_state(state):
    """Communities state."""
    state.COMMUNITIES.add(
        "comm",
        {
            "id": "12345678-abcd-1a2b-3c4d-123abc456def",
            "bucket_id": "12345678-abcd-1a2b-3c4d-123abc456def",
            "oai_set_id": 1,
            "community_file_id": None,
            "logo_object_version_id": None,
        },
    )

    return state.COMMUNITIES


@pytest.fixture(scope="function")
def secret_keys_state(state):
    """Adds secret keys to global state."""
    state.VALUES.add(
        "old_secret_key",
        {"value": bytes("OLDKEY", "utf-8")},
    )
    state.VALUES.add(
        "new_secret_key",
        {"value": bytes("NEWKEY", "utf-8")},
    )
    return state


@pytest.fixture(scope="function")
def buckets_state(state):
    """Adds a bucket to draft map to the state."""
    state.BUCKETS.add(
        "0e12b4b6-9cc7-46df-9a04-c11c478de211",
        {"draft_id": "d94f793c-47d2-48e2-9867-ca597b4ebb41"},
    )
    return state



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration record table load module."""

from ....load.postgresql.bulk.generators import TableGenerator
from ....state import STATE
from ...models.records import RDMVersionState


class RDMVersionStateTableGenerator(TableGenerator):
    """RDM version state computed table."""

    def __init__(self):
        """Constructor."""
        super().__init__(tables=[RDMVersionState])

    def _generate_rows(self, parent_entry, **kwargs):
        # Version state to be populated in the end from the final state
        yield RDMVersionState(
            latest_index=parent_entry.get("latest_index"),
            parent_id=parent_entry["id"],
            latest_id=parent_entry.get("latest_id"),
            next_draft_id=parent_entry.get("next_draft_id"),
        )

    def prepare(self, tmp_dir, entry, stack, output_files, **kwargs):
        """Compute rows."""
        # overwrite to avoid it using the parent class on a per
        pass

    def post_prepare(self, tmp_dir, stack, output_files, **kwargs):
        """Overwrite entries with parent state entries."""
        for entry in STATE.PARENTS.all():
            super().prepare(tmp_dir, entry, stack, output_files, **kwargs)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration users stream."""

from .load import CommunityCopyLoad
from .transform import (
    CommunityEntry,
    CommunityFileEntry,
    CommunityMemberEntry,
    CommunityTransform,
    FeaturedCommunityEntry,
    OAISetEntry,
    ParentCommunityEntry,
)

__all__ = (
    "CommunityCopyLoad",
    "CommunityEntry",
    "CommunityMemberEntry",
    "CommunityTransform",
    "CommunityFileEntry",
    "FeaturedCommunityEntry",
    "OAISetEntry",
    "ParentCommunityEntry",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Base table generator."""

import csv
from dataclasses import fields
from datetime import datetime
from uuid import UUID

import orjson

from ...generators import PostgreSQLGenerator


def as_csv_row(dc):
    """Serialize a dataclass instance as a CSV-writable row."""
    row = []
    for f in fields(dc):
        val = getattr(dc, f.name)
        if val:
            if issubclass(f.type, (dict,)):
                val = orjson.dumps(val).decode("utf-8")
            elif issubclass(f.type, (datetime,)):
                val = val.isoformat()
            elif issubclass(f.type, (UUID,)):
                val = str(val)
        row.append(val)
    return row


class TableGenerator(PostgreSQLGenerator):
    """Create CSV files with table create and inserts."""

    def __init__(self, tables, pks=None, post_load_hooks=None, existing_data=False):
        """Constructor."""
        super().__init__(pks, post_load_hooks)
        self.tables = tables
        self.existing_data = existing_data

    def prepare(self, tmp_dir, entry, stack, output_files, create=False, **kwargs):
        """Compute rows."""
        if not self.existing_data:
            # is_db_empty would come in play and make _generate_pks optional
            self._generate_pks(entry, create)
            # resolve entry references
            self._resolve_references(entry)

            for entry in self._generate_rows(entry):
                if entry.__tablename__ not in output_files:
                    fpath = tmp_dir / f"{entry.__tablename__}.csv"
                    output_files[entry.__tablename__] = csv.writer(
                        stack.enter_context(open(fpath, "w+"))
                    )
                writer = output_files[entry.__tablename__]
                writer.writerow(as_csv_row(entry))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration parent table load module."""

from datetime import datetime

from invenio_rdm_migrator.load.ids import pid_pk

from ...models.pids import PersistentIdentifier
from ...models.records import RDMParentMetadata


def generate_parent_rows(parent):
    """Generates rows for a parent record."""
    now = datetime.utcnow().isoformat()
    parent_pid = parent["json"]["pid"]
    # order is important when doing action/streaming migration
    # parent recid
    yield PersistentIdentifier(
        id=parent_pid["pk"],
        pid_type=parent_pid["pid_type"],
        pid_value=parent["json"]["id"],
        status=parent_pid["status"],
        object_type=parent_pid["obj_type"],
        object_uuid=parent["id"],
        created=now,
        updated=now,
    )
    # parent DOI
    parent_doi = parent["json"].get("pids", {}).get("doi")
    if parent_doi and parent_doi["identifier"]:
        yield PersistentIdentifier(
            id=pid_pk(),
            pid_type="doi",
            pid_value=parent_doi["identifier"],
            status="R",
            object_type="rec",
            object_uuid=parent["id"],
            created=now,
            updated=now,
        )

    # parent record
    yield RDMParentMetadata(
        id=parent["id"],
        json=parent["json"],
        created=parent["created"],
        updated=parent["updated"],
        version_id=parent["version_id"],
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration drafts row load module."""

from dataclasses import dataclass
from datetime import datetime
from typing import Optional

from ....actions import LoadAction, LoadData
from ....load.ids import generate_pk, generate_uuid
from ....load.postgresql.transactions.operations import Operation, OperationType
from ....state import STATE
from ...models.files import FilesBucket
from ...models.pids import PersistentIdentifier
from ...models.records import (
    RDMDraftFile,
    RDMDraftMetadata,
    RDMParentMetadata,
    RDMRecordFile,
    RDMRecordMetadata,
    RDMVersionState,
)
from ...records.table_generators.references import (
    CommunitiesReferencesMixin,
    PIDsReferencesMixin,
)
from .parents import generate_parent_ops


@dataclass
class RDMDraftCreateData(LoadData):
    """Draft create action data."""

    parent_pid: dict
    parent: dict
    draft_pid: dict
    draft: dict
    draft_bucket: dict


class DraftCreateAction(LoadAction, CommunitiesReferencesMixin, PIDsReferencesMixin):
    """RDM draft creation."""

    name = "create-draft"
    data_cls = RDMDraftCreateData

    pks = [
        # not both parent.json.pid and draft.json.pid come filled in from the
        # transform action to match parent_pid and draft_pid data.
        ("draft_pid", "id", generate_pk),
        ("draft", "id", generate_uuid),
        ("parent_pid", "id", generate_pk),
        ("parent", "id", generate_uuid),
    ]

    def _generate_rows(self, **kwargs):
        """Generates rows for a new draft."""
        yield from self._generate_pid_rows(**kwargs)
        yield from self._generate_bucket_rows(**kwargs)
        yield from self._generate_draft_rows(**kwargs)

    def _generate_pid_rows(self, **kwargs):
        """Generates rows for a new draft."""
        pid = self.data.draft_pid
        if pid["pid_type"] != "depid":
            # note would raise an exception if it exists
            STATE.PIDS.add(
                pid["pid_value"],  # recid
                {
                    "id": pid["id"],
                    "pid_type": pid["pid_type"],
                    "status": pid["status"],
                    "obj_type": pid["object_type"],
                    "created": pid["created"],
                },
            )
            yield Operation(OperationType.INSERT, PersistentIdentifier, pid)

    def _generate_bucket_rows(self, **kwargs):
        """Generates rows for a new draft."""
        yield Operation(OperationType.INSERT, FilesBucket, self.data.draft_bucket)

    def _generate_draft_rows(self, **kwargs):
        """Generates rows for a new draft."""
        now = datetime.utcnow().isoformat()

        draft = self.data.draft
        parent = self.data.parent

        forked_published = STATE.RECORDS.get(draft["json"]["id"])
        existing_parent = STATE.PARENTS.get(parent["json"]["id"])

        # parent id
        #  a) draft of a published record, parent id = parent id of published
        #  b) new version, parent id = parent id of the previous version
        #  c) draft of a new record, parent id = given by pk func

        # both values should be equal at first, the cannot be set in the transform step
        # parent.id is calculated in the pks step
        draft["parent_id"] = parent["id"]
        if not existing_parent:  # case c
            STATE.PARENTS.add(
                parent["json"]["id"],  # recid
                {"id": parent["id"], "next_draft_id": draft["id"]},
            )
            # drafts have a parent on save
            # on the other hand there is no community parent/request
            yield from generate_parent_ops(parent, self.data.parent_pid)

        else:  # case a and b
            parent["id"] = existing_parent["id"]
            draft["parent_id"] = existing_parent["id"]  # keep metadata consistent
            if not forked_published:
                # it can only happen once
                assert not existing_parent.get("next_draft_id")
                STATE.PARENTS.update(
                    parent["json"]["id"],
                    {"next_draft_id": draft["id"]},
                )
            else:
                # state parent  and an existing record must match
                assert parent["id"] == forked_published["parent_id"]

        if not forked_published:
            # recid must have been created by a previous action in the same tx group
            draft_pid = STATE.PIDS.get(draft["json"]["id"])
            assert draft_pid

            # update to add object_uuid
            # could avoid this operation but it is clearer on when and why this happens
            yield Operation(
                OperationType.UPDATE,
                PersistentIdentifier,
                dict(
                    id=draft_pid["id"],  # pk
                    pid_type=draft_pid["pid_type"],  # in drafts are recid
                    pid_value=draft["json"]["id"],
                    status=draft_pid["status"],
                    object_type="rec",  # hardcoded since the state has the initial one
                    object_uuid=draft["id"],
                    created=draft_pid["created"],
                    updated=now,
                ),
            )

        draft_id = forked_published.get("id") or draft["id"]

        STATE.BUCKETS.add(draft["bucket_id"], {"draft_id": draft_id})
        yield Operation(
            OperationType.INSERT,
            RDMDraftMetadata,
            dict(
                id=draft_id,
                json=draft["json"],
                created=draft["created"],
                updated=draft["updated"],
                version_id=draft["version_id"],
                index=forked_published.get("index") or draft["index"],
                bucket_id=draft["bucket_id"],
                parent_id=parent["id"],
                expires_at=draft["expires_at"],
                fork_version_id=forked_published.get("fork_version_id")
                or draft["fork_version_id"],
            ),
        )

        # this query can be avoided by keeping a consistent view across this method
        existing_parent = STATE.PARENTS.get(parent["json"]["id"])
        version_op = OperationType.UPDATE if forked_published else OperationType.INSERT
        yield Operation(
            version_op,
            RDMVersionState,
            dict(
                latest_index=existing_parent["latest_index"],
                parent_id=existing_parent["id"],
                latest_id=existing_parent["latest_id"],
                next_draft_id=existing_parent["next_draft_id"],
            ),
        )

    def _resolve_references(self, **kwargs):
        """Resolve references e.g communities slug names."""
        # resolve parent communities slug
        parent = self.data.parent
        communities = parent["json"].get("communities")
        if communities:
            self.resolve_communities(communities)
        self.resolve_draft_pids(self.data.draft)


@dataclass
class RDMDraftEditData(LoadData):
    """Draft edit action data."""

    draft: dict
    parent: dict


class DraftEditAction(LoadAction, CommunitiesReferencesMixin, PIDsReferencesMixin):
    """RDM draft edit/update."""

    name = "edit-draft"
    data_cls = RDMDraftEditData

    def _generate_rows(self, **kwargs):
        """Generates rows for a new draft."""
        draft = self.data.draft
        parent = self.data.parent

        assert parent["id"]  # make sure we have an id to update on
        # data and model keys/fields are the same
        yield Operation(OperationType.UPDATE, RDMParentMetadata, parent)

        forked_published = STATE.RECORDS.get(draft["json"]["id"])
        draft_id = forked_published.get("id") or draft["id"]
        assert draft_id
        draft["id"] = draft_id

        # the index and forked_version_id have been properly set from cached when the draft
        # was created, therefore now they are only passed (from the draft) if given in the
        # partial data
        # the parent_id cannot change unless is a support operation (e.g. merge) which
        # will not happen during migration
        # therefore all fields come from the draft and match data/model keys/fields
        yield Operation(OperationType.UPDATE, RDMDraftMetadata, draft)

    def _resolve_references(self, **kwargs):
        """Resolve references e.g communities slug names."""
        # resolve parent communities slug
        parent = self.data.parent
        communities = parent["json"].get("communities")
        if communities:
            self.resolve_communities(communities)
        self.resolve_draft_pids(self.data.draft)

        # resolve parent and draft uuid from versioning table
        state_parent = STATE.PARENTS.get(parent["json"]["id"])
        self.data.draft["id"] = state_parent["next_draft_id"]
        self.data.parent["id"] = state_parent["id"]


@dataclass
class RDMDraftPublishData(LoadData):
    """Draft publish action data."""

    bucket: dict
    parent: dict
    draft: dict
    parent_pid: dict
    draft_pid: dict
    draft_oai: dict
    parent_doi: Optional[dict] = None
    draft_doi: Optional[dict] = None


class DraftPublishAction(LoadAction, CommunitiesReferencesMixin, PIDsReferencesMixin):
    """RDM publish creation."""

    # IMPORTANT: at the moment it assumes publishing a new draft
    # not taking into account forked_published or other cases

    name = "publish-draft"
    data_cls = RDMDraftPublishData

    def _generate_rows(self, **kwargs):
        """Generates rows for a new draft."""
        yield from self._generate_pid_rows(**kwargs)
        yield from self._generate_bucket_rows(**kwargs)
        yield from self._generate_draft_rows(**kwargs)

    def _generate_pid_rows(self, **kwargs):
        """Generates rows for a new draft."""
        # recid (draft) gets registered when creating a record, keeps the same pid
        assert self.data.draft_pid["status"] == "R"
        assert self.data.parent_pid["status"] == "R"
        # no need to update parent pids, they are not in the state
        STATE.PIDS.update(self.data.draft_pid["pid_value"], {"status": "R"})

        yield Operation(OperationType.UPDATE, PersistentIdentifier, self.data.draft_pid)
        yield Operation(
            OperationType.UPDATE, PersistentIdentifier, self.data.parent_pid
        )

        if self.data.draft_doi:
            assert self.data.parent_doi  # cannot be one doi without the other
            yield Operation(
                OperationType.INSERT, PersistentIdentifier, self.data.draft_doi
            )
            yield Operation(
                OperationType.INSERT, PersistentIdentifier, self.data.parent_doi
            )

        if self.data.draft_oai:
            yield Operation(
                OperationType.INSERT, PersistentIdentifier, self.data.draft_oai
            )

    def _generate_bucket_rows(self, **kwargs):
        """Generates rows for a new draft."""
        # the drafts service would make a copy of all object versions to the copied bucket
        # however, at DB level updating the bucket information would suffice
        # the new record would point to it and the previous draft would be deleted.
        assert self.data.bucket["locked"]
        # delete bucket from STATE.BUCKET
        # it is not linked to a draft anymore and records cannot get file updates
        STATE.BUCKETS.delete(self.data.bucket["id"])
        # need to update the bucket (e.g. for locking it)
        yield Operation(OperationType.UPDATE, FilesBucket, self.data.bucket)

        # TODO: impelement search on the state
        files = STATE.FILE_RECORDS.search("record_id", self.data.draft["id"])
        for file in files:
            yield Operation(OperationType.DELETE, RDMDraftFile, file)
            yield Operation(OperationType.INSERT, RDMRecordFile, file)
            STATE.FILE_RECORDS.delete(file["id"])

    def _generate_draft_rows(self, **kwargs):
        """Generates rows for a new draft."""
        parent = self.data.parent
        draft = self.data.draft
        # delete draft
        # can we simplify this by only passing an `id`?
        # would mean all other fields are optional
        # i.e. data validation is delegated to db exceptions
        yield Operation(
            OperationType.DELETE,
            RDMDraftMetadata,
            dict(
                id=draft["id"],
                json=draft["json"],
                created=draft["created"],
                updated=draft["updated"],
                version_id=draft["version_id"],
                index=draft["index"],
                bucket_id=draft["bucket_id"],
                parent_id=parent["id"],
                expires_at=draft["expires_at"],
                fork_version_id=draft["fork_version_id"],
            ),
        )

        # create record
        yield Operation(
            OperationType.INSERT,
            RDMRecordMetadata,
            dict(
                id=draft["id"],
                json=draft["json"],  # what happens with fields such as $schema?
                created=draft["updated"],  # creation of record ~ latest draft update
                updated=draft["updated"],
                version_id=1,
                index=1,
                bucket_id=draft["bucket_id"],
                parent_id=parent["id"],
            ),
        )
        # update parent
        yield Operation(
            OperationType.UPDATE,
            RDMParentMetadata,
            dict(
                id=parent["id"],
                json=parent["json"],
                created=parent["created"],
                updated=parent["updated"],
                version_id=parent["version_id"],
            ),
        )

        # update versioning
        STATE.PARENTS.update(
            parent["json"]["id"],
            {
                "id": parent["id"],
                "latest_id": draft["id"],
                "latest_index": draft["index"],
                "next_draft_id": None,
            },
        )
        yield Operation(
            OperationType.UPDATE,
            RDMVersionState,
            dict(
                parent_id=parent["id"],
                latest_id=draft["id"],
                latest_index=draft["index"],
                next_draft_id=None,  # publishing a new record
            ),
        )

    def _resolve_references(self, **kwargs):
        """Resolve references e.g communities slug names."""
        # TODO: dup code, move to base DraftAction class?
        # resolve parent communities slug
        parent = self.data.parent
        communities = parent["json"].get("communities")
        if communities:
            self.resolve_communities(communities)
        self.resolve_draft_pids(self.data.draft)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Datetime transformation tests."""

from datetime import datetime

from invenio_rdm_migrator.transform import DatetimeMixin


class TestTransform(DatetimeMixin):
    """Test Datetime transform."""

    pass


def test_date_microseconds():
    data = {"one": 1620000000000000, "two": 1620000000000000, "three": 1620000000000000}

    TestTransform()._microseconds_to_isodate(data=data, fields=["one", "two"])

    assert data == {
        "one": "2021-05-03T00:00:00",
        "two": "2021-05-03T00:00:00",
        "three": 1620000000000000,  # left untouched
    }


def test_date_microseconds_no_int():
    data = {"one": "1620000000000000"}

    TestTransform()._microseconds_to_isodate(data=data, fields=["one"])

    assert data == {"one": "1620000000000000"}


def test_date_milliseconds():
    data = {"one": 1620000000000, "two": 1620000000000, "three": 1620000000000}

    TestTransform()._milliseconds_to_isodate(data=data, fields=["one", "two"])

    assert data == {
        "one": "2021-05-03T00:00:00",
        "two": "2021-05-03T00:00:00",
        "three": 1620000000000,  # left untouched
    }


def test_date_milliseconds_no_int():
    data = {"one": "1620000000000"}

    TestTransform()._milliseconds_to_isodate(data=data, fields=["one"])

    assert data == {"one": "1620000000000"}



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration transform interfaces."""

from sqlalchemy_utils.types.encrypted.encrypted_type import AesEngine

from ..state import STATE


class EncryptMixin:
    """Enables re-encryption of values based on old and new secret keys."""

    def __init__(self, *args, **kwargs):
        """Constructor."""
        state = STATE.VALUES
        old_secret_key = state.get("old_secret_key")["value"]
        assert old_secret_key
        self.decrypt_engine = self._init_engine(old_secret_key)

        new_secret_key = state.get("new_secret_key")["value"]
        assert new_secret_key
        self.encrypt_engine = self._init_engine(new_secret_key)
        super().__init__(*args, **kwargs)

    @staticmethod
    def _init_engine(secret_key):
        """Initialize an encryption engine."""
        engine = AesEngine()
        engine._update_key(secret_key)
        engine._set_padding_mechanism()

        return engine

    def re_encrypt(self, value):
        """Value re-encryption.

        Following invenio_db.utils:rebuild_encrypted_properties.
        This is the function used by invenio-oauthclient/oauth2server to re-encrypt
        when the secret key changes.
        It boils doen to sqlalchemy-utils AESEngine encrypt/decrypt.
        """
        decrypted_token = self.decrypt_engine.decrypt(value)
        new_value = self.encrypt_engine.encrypt(decrypted_token)

        return new_value



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration transaction extract classes."""

from dataclasses import dataclass


@dataclass
class Tx:
    """An extracted DB transaction."""

    id: int
    operations: list[dict]  # TODO: we could more narrowly define it later



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""PostgreSQL transaction load tests."""

from dataclasses import dataclass
from typing import Optional

import pytest
from sqlalchemy.orm import Mapped, mapped_column

from invenio_rdm_migrator.actions import LoadAction, LoadData
from invenio_rdm_migrator.load.postgresql.models import Model
from invenio_rdm_migrator.load.postgresql.transactions import PostgreSQLTx
from invenio_rdm_migrator.load.postgresql.transactions.operations import (
    Operation,
    OperationType,
)


@dataclass
class TransactionData(LoadData):
    """Test load data."""

    id: int
    test: Optional[str] = None


class TestLoadAction(LoadAction):
    """Test load action."""

    name = "test-transaction-load"
    data_cls = TransactionData

    def _generate_rows(self, **kwargs):  # pragma: no cover
        """Yield generated rows."""
        op_type = OperationType.INSERT
        if not self.data.test:
            op_type = OperationType.DELETE
        elif "update" in self.data.test:
            op_type = OperationType.UPDATE

        yield Operation(
            type=op_type,
            model=TestModel,
            data={"id": self.data.id, "test": self.data.test},
        )


class TestModel(Model):
    """Persistent identifier dataclass model."""

    __tablename__ = "transaction-test"

    id: Mapped[int] = mapped_column(primary_key=True)
    test: Mapped[str]


@pytest.fixture(scope="module")
def database(engine):
    tables = [TestModel]

    # create tables
    for model in tables:
        model.__table__.create(bind=engine, checkfirst=True)

    yield engine

    # remove tables
    for model in tables:
        model.__table__.drop(engine)


@pytest.fixture(scope="function")
def test_item(database, session):
    obj = TestModel(id=1, test="test item")
    session.add(obj)
    session.commit()


@pytest.fixture(scope="function")
def pg_tx(db_uri, session):
    return PostgreSQLTx(db_uri=db_uri, _session=session)


###
# Operation Type
###


def test_load_transaction_insert(session, pg_tx, database):
    tx = [TestLoadAction(data={"tx_id": 1, "id": 101, "test": "insert test"})]
    pg_tx.run(tx)

    result = session.query(TestModel).all()
    assert len(result) == 1
    assert result[0].id == 101
    assert result[0].test == "insert test"

    # suboptimal but pragmatic cleanup
    session.delete(result[0])
    session.commit()


def test_load_transaction_delete(pg_tx, session, database, test_item):
    tx = [TestLoadAction(data={"tx_id": 2, "id": 1})]
    pg_tx.run(tx)

    session.query(TestModel).all()


def test_load_transaction_update(pg_tx, session, database, test_item):
    tx = [TestLoadAction(data={"tx_id": 1, "id": 1, "test": "update test"})]
    pg_tx.run(tx)

    result = session.query(TestModel).all()
    assert len(result) == 1
    assert result[0].id == 1
    assert result[0].test == "update test"


# ###
# # Transaction cardinality
# ###


def test_load_no_transaction(pg_tx, session, database):
    pg_tx.run([])

    result = session.query(TestModel).all()
    assert len(result) == 0


def test_load_transaction_multiple_actions(pg_tx, session, database):
    tx = [
        TestLoadAction(data={"tx_id": 1, "id": 101, "test": "first instance"}),
        TestLoadAction(data={"tx_id": 2, "id": 102, "test": "second instance"}),
    ]
    pg_tx.run(tx)

    result = session.query(TestModel).all()
    assert len(result) == 2
    assert result[0].id == 101
    assert result[0].test == "first instance"
    assert result[1].id == 102
    assert result[1].test == "second instance"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""PostgreSQL bulk load tests."""

import tempfile
from dataclasses import InitVar
from unittest.mock import patch

import pytest
from sqlalchemy.orm import Mapped, mapped_column

from invenio_rdm_migrator.load.postgresql.bulk import PostgreSQLCopyLoad
from invenio_rdm_migrator.load.postgresql.bulk.generators import (
    ExistingDataTableGenerator,
    SingleTableGenerator,
)
from invenio_rdm_migrator.load.postgresql.models import Model


@pytest.fixture(scope="function")
def data_dir():
    """Yields a temporary directory."""
    data_dir = tempfile.TemporaryDirectory()
    yield data_dir
    data_dir.cleanup()


###
#  Table Generators
###


class TestModel(Model):
    """Test dataclass model."""

    foo: Mapped[str]
    bar: Mapped[str]
    number: Mapped[int] = mapped_column(primary_key=True)

    __tablename__: InitVar[str] = "test_table"


def test_identity_tg_generate_rows():
    input = {"foo": "some", "bar": "value", "number": 10}
    expected = TestModel(foo="some", bar="value", number=10)

    assert next(SingleTableGenerator(table=TestModel)._generate_rows(input)) == expected


def test_identity_tg_multiple_tables():
    with pytest.raises(AssertionError):
        SingleTableGenerator(table=[TestModel, TestModel])


###
# Existing data
###


class TestModelToo(Model):
    """Another dataclass model."""

    foo: Mapped[str]
    bar: Mapped[str]
    number: Mapped[int] = mapped_column(primary_key=True)

    __tablename__: InitVar[str] = "test_table_too"


class CopyLoad(PostgreSQLCopyLoad):
    """Mock load class."""

    def __init__(self, **kwargs):
        """Constructor."""
        super().__init__(
            table_generators=[
                ExistingDataTableGenerator(tables=[TestModelToo]),
                SingleTableGenerator(table=TestModel),
            ],
            **kwargs
        )


@pytest.fixture(scope="function")
def entries():
    return [{"foo": "test", "bar": "also testing", "number": 1}]


def mock_load(_, table_entries):
    """Mock method to assert the iterators content.

    Otherwise the assert would check memory addresses which do not match.
    """
    assert list(table_entries) == [(True, TestModelToo), (True, TestModel)]


@patch.object(CopyLoad, "_load", mock_load)
@patch.object(CopyLoad, "_post_load")  # needs mocking due to db connection
@patch.object(ExistingDataTableGenerator, "prepare")
@patch.object(SingleTableGenerator, "prepare")
def test_load_with_all_existing_data(
    m_prepare_i, m_prepare_ii, _, data_dir, tmp_dir, entries
):
    """Checks the load does not call prepare on the tg (i.e. no data written)."""
    load = CopyLoad(
        db_uri=None, data_dir=data_dir.name, tmp_dir=tmp_dir.name, existing_data=True
    )
    load.run(entries)

    # the table generators prepare step (write csv files) were not called
    m_prepare_i.assert_not_called()
    m_prepare_ii.assert_not_called()

    # check the load would be done with all existing data is done via mock_load


def mock_load(_, table_entries):
    """Mock method to assert the iterators content.

    Otherwise the assert would check memory addresses which do not match.
    """
    # note the second item on the list has a False as existing_data
    assert list(table_entries) == [(True, TestModelToo), (False, TestModel)]


@patch.object(CopyLoad, "_load", mock_load)
@patch.object(CopyLoad, "_post_load")  # needs mocking due to db connection
@patch.object(ExistingDataTableGenerator, "_generate_rows")
@patch.object(SingleTableGenerator, "_generate_rows")
def test_load_with_existing_data_tg(
    m_gen_rows_i, m_gen_rows_ii, _, data_dir, tmp_dir, entries
):
    """Checks the load calls prepare on one tg."""
    load = CopyLoad(db_uri=None, data_dir=data_dir.name, tmp_dir=tmp_dir.name)
    load.run(entries)

    # the `prepare` method is called on both tg, however the filtering happens at tg level
    # therefore we assert on row generation (could also be done on pk or refs)
    m_gen_rows_i.assert_called()
    m_gen_rows_ii.assert_not_called()  # not called

    # check the load would be done with all existing data is done via mock_load


class CopyLoadToo(PostgreSQLCopyLoad):
    """Mock load class."""

    def __init__(self, **kwargs):
        """Constructor."""
        super().__init__(
            table_generators=[
                SingleTableGenerator(table=TestModelToo),
                SingleTableGenerator(table=TestModel),
            ],
            **kwargs
        )


def mock_load(_, table_entries):
    """Mock method to assert the iterators content.

    Otherwise the assert would check memory addresses which do not match.
    """
    assert list(table_entries) == [(False, TestModelToo), (False, TestModel)]


@patch.object(CopyLoadToo, "_load", mock_load)
@patch.object(CopyLoadToo, "_post_load")  # needs mocking due to db connection
@patch.object(SingleTableGenerator, "_generate_rows")
def test_load_without_existing_data(m_gen_rows, _, data_dir, tmp_dir, entries):
    """Checks the load calls prepare on both tg."""

    load = CopyLoadToo(db_uri=None, data_dir=data_dir.name, tmp_dir=tmp_dir.name)
    load.run(entries)

    # the `prepare` method is called on both tg, however the filtering happens at tg level
    # therefore we assert on row generation (could also be done on pk or refs)
    m_gen_rows.call_count = 2

    # check the load would be done with all existing data is done via mock_load



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Identity transformation tests."""

import pytest

from invenio_rdm_migrator.transform import (
    IdentityDictKeyMixin,
    IdentityTransform,
    Transform,
)


@pytest.fixture(scope="module")
def transform_with_mixin():
    """Test instance of a transform class with identity mixin"""

    class FixtureTransform(Transform, IdentityDictKeyMixin):
        """Transform fixture class."""

        def _implemented(self, entry):
            return "foo"

        def _transform(self, entry):
            """Transform entry."""
            return {
                "implemented": self._implemented(entry),
                "unimplemented": self._unimplemented(entry),
            }

    return FixtureTransform()


def test_identity_dict_key_mixin(transform_with_mixin):
    input = {"implemented": "test", "unimplemented": "test too"}
    expected = {"implemented": "foo", "unimplemented": "test too"}

    # assert the implemented method has priority over the dictionary value
    # assert the unimplemented method falls back to the dictionary value
    assert transform_with_mixin._transform(input) == expected


def test_identity_transform():
    input = {"implemented": "test", "unimplemented": "test too"}
    expected = {"implemented": "test", "unimplemented": "test too"}

    assert IdentityTransform()._transform(input) == expected



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration user table load module."""


from ...load.postgresql.bulk.generators import TableGenerator
from ..models.users import LoginInformation, User, UserIdentity


class UserTableGenerator(TableGenerator):
    """User and related tables load."""

    def __init__(self):
        """Constructor."""
        super().__init__(
            tables=[
                User,
                LoginInformation,
                UserIdentity,
            ],
        )

    def _generate_rows(self, data, **kwargs):
        user = data["user"]
        login_info = user.pop("login_information", None)
        yield User(**user)
        if login_info:
            yield LoginInformation(
                user_id=user["id"],
                **login_info,
            )
        identities = data.get("identities", [])
        for identity in identities:
            yield UserIdentity(
                id_user=user["id"],
                **identity,
            )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""InvenioRDM migration streams runner."""

from pathlib import Path

import yaml

from ..logging import Logger
from ..state import STATE, StateDB
from .records.state import ParentModelValidator
from .streams import Stream


class Runner:
    """ETL streams runner."""

    def _read_config(self, filepath):
        """Read config from file."""
        with open(filepath) as f:
            return yaml.safe_load(f)

    def __init__(self, stream_definitions, config_filepath):
        """Constructor."""
        config = self._read_config(config_filepath)

        self.data_dir = Path(config.get("data_dir"))
        self.data_dir.mkdir(parents=True, exist_ok=True)

        self.tmp_dir = Path(config.get("tmp_dir"))
        self.tmp_dir.mkdir(parents=True, exist_ok=True)

        self.state_dir = Path(config.get("state_dir"))
        self.state_dir.mkdir(parents=True, exist_ok=True)

        self.log_dir = Path(config.get("log_dir"))
        self.log_dir.mkdir(parents=True, exist_ok=True)

        Logger.initialize(self.log_dir)

        self.db_uri = config.get("db_uri")
        self.streams = []
        self.state = StateDB(
            db_dir=self.state_dir, validators={"parents": ParentModelValidator}
        )
        STATE.initialized_state(self.state)

        # set up secret keys
        for key in ("old_secret_key", "new_secret_key"):
            stored_value = STATE.VALUES.get(key)
            if stored_value:
                STATE.VALUES.update(key, {"value": bytes(config.get(key), "utf-8")})
            else:
                STATE.VALUES.add(key, {"value": bytes(config.get(key), "utf-8")})

        # start processing streams
        for definition in stream_definitions:
            if definition.name in config:
                # get will return a None for e.g. files:
                stream_config = config.get(definition.name) or {}
                # merge state objects from stream definition config
                existing_data = stream_config.get("existing_data", {})

                # if loading pass source data dir, else pass tmp to dump new csv files
                data_dir = self.data_dir
                tmp_dir = self.tmp_dir / definition.name
                extract = None
                transform = None
                if not existing_data:
                    extract = definition.extract_cls(**stream_config.get("extract", {}))
                    transform = definition.transform_cls(
                        **stream_config.get("transform", {})
                    )

                self.streams.append(
                    Stream(
                        definition.name,
                        extract,
                        transform,
                        definition.load_cls(
                            db_uri=self.db_uri,
                            data_dir=data_dir,
                            tmp_dir=tmp_dir,
                            existing_data=existing_data,
                            **stream_config.get("load", {}),
                        ),
                    )
                )

    def run(self):
        """Run ETL streams."""
        for stream in self.streams:
            try:
                stream.run()
                # on successful stream run, persist state
                self.state.save(filename=f"{stream.name}.db")
            except Exception:
                Logger.get_logger().exception(
                    f"Stream {stream.name} failed.", exc_info=1
                )
                continue



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Action streams test configuration."""

import pytest


###
# Draft
###
@pytest.fixture(scope="function")
def draft_pid_data():
    """PID data."""
    return {
        "id": 12132090,
        "pid_type": "recid",
        "pid_value": "1217215",
        "pid_provider": None,
        "status": "K",
        "object_type": "rec",
        "object_uuid": "d94f793c-47d2-48e2-9867-ca597b4ebb41",
        "created": "2022-01-01T00:00:00",
        "updated": "2022-01-01T00:00:00",
    }


@pytest.fixture(scope="function")
def draft_doi_data():
    """PID data."""
    return {
        "id": 12132092,
        "pid_type": "doi",
        "pid_value": "10.12345/rdm.1217215",
        "pid_provider": None,
        "status": "K",
        "object_type": "rec",
        "object_uuid": "d94f793c-47d2-48e2-9867-ca597b4ebb41",
        "created": "2022-01-01T00:00:00",
        "updated": "2022-01-01T00:00:00",
    }


@pytest.fixture(scope="function")
def draft_data():
    """Draft data."""
    return {
        "id": "d94f793c-47d2-48e2-9867-ca597b4ebb41",
        "json": {
            "id": "1217215",
            "$schema": "https://zenodo.org/schemas/deposits/records/record-v1.0.0.json",
            "pids": {},
            "files": {"enabled": True},
            "metadata": {},
            "access": {
                "record": "public",
                "files": "public",
            },
            "custom_fields": {},
        },
        "version_id": 1,
        "index": 1,
        "bucket_id": "0e12b4b6-9cc7-46df-9a04-c11c478de211",
        "parent_id": "9493793c-47d2-48e2-9867-ca597b4ebb42",
        "expires_at": None,
        "fork_version_id": None,
        "created": "2021-05-01T00:00:00",
        "updated": "2021-05-01T00:00:00",
    }


###
# Parent
###


@pytest.fixture(scope="function")
def parent_pid_data():
    """PID data."""
    return {
        "id": 12132091,
        "pid_type": "recid",
        "pid_value": "1217214",
        "pid_provider": None,
        "status": "R",
        "object_type": "rec",
        "object_uuid": "d94f793c-47d2-48e2-9867-ca597b4ebb42",
        "created": "2022-01-01T00:00:00",
        "updated": "2022-01-01T00:00:00",
    }


@pytest.fixture(scope="function")
def parent_doi_data():
    """PID data."""
    return {
        "id": 12132093,
        "pid_type": "doi",
        "pid_value": "10.12345/rdm.1217214",
        "pid_provider": None,
        "status": "K",
        "object_type": "rec",
        "object_uuid": "d94f793c-47d2-48e2-9867-ca597b4ebb41",
        "created": "2022-01-01T00:00:00",
        "updated": "2022-01-01T00:00:00",
    }


@pytest.fixture(scope="function")
def parent_data():
    """Parent data."""
    return {
        "id": "9493793c-47d2-48e2-9867-ca597b4ebb42",
        "json": {
            "id": "1217214",
            "pid": {"pk": 2, "pid_type": "recid", "status": "R", "obj_type": "rec"},
            "access": {"owned_by": {"user": 1234}},
            "communities": {"ids": ["zenodo", "migration"], "default": "zenodo"},
        },
        "version_id": 1,
        "created": "2021-05-01T00:00:00",
        "updated": "2021-05-01T00:00:00",
    }


###
# Record
###


@pytest.fixture(scope="function")
def record_oai_data():
    """PID data."""
    return {
        "id": 12132094,
        "pid_type": "oai",
        "pid_value": "oai:zenodo.org:1217215",
        "pid_provider": None,
        "status": "K",
        "object_type": "rec",
        "object_uuid": "d94f793c-47d2-48e2-9867-ca597b4ebb41",
        "created": "2022-01-01T00:00:00",
        "updated": "2022-01-01T00:00:00",
    }


###
# Files
###


@pytest.fixture(scope="function")
def bucket_data():
    """Bucket data."""
    return {
        "id": "0e12b4b6-9cc7-46df-9a04-c11c478de211",
        "created": "2022-01-01T00:00:00",
        "updated": "2022-01-01T00:00:00",
        "default_location": 1,
        "default_storage_class": "L",
        "size": 0,
        "quota_size": 50000000000,
        "max_file_size": 50000000000,
        "locked": False,
        "deleted": False,
    }


@pytest.fixture(scope="function")
def bucket_data_w_size():
    """Bucket data."""
    return {
        "id": "0e12b4b6-9cc7-46df-9a04-c11c478de211",
        "created": "2023-06-29T13:00:00",
        "updated": "2023-06-29T14:00:00",
        "default_location": 1,
        "default_storage_class": "S",
        "size": 1562554,
        "quota_size": 50000000000,
        "max_file_size": 50000000000,
        "locked": False,
        "deleted": False,
    }


@pytest.fixture(scope="function")
def ov_data():
    """Object version data."""
    return {
        "version_id": "f8200dc7-55b6-4785-abd0-f3d13b143c98",
        "created": "2023-06-29T13:00:00",
        "updated": "2023-06-29T14:00:00",
        "key": "IMG_3535.jpg",
        "bucket_id": "0e12b4b6-9cc7-46df-9a04-c11c478de211",
        "file_id": "e94b243e-9c0c-44df-bd1f-6decc374cf78",
        "_mimetype": None,
        "is_head": True,
    }


@pytest.fixture(scope="function")
def fi_data():
    """File instance data."""
    return {
        "id": "e94b243e-9c0c-44df-bd1f-6decc374cf78",
        "created": "2023-06-29T13:00:00",
        "updated": "2023-06-29T14:00:00",
        "uri": "root://eosmedia.cern.ch//eos/media/zenodo/test/data/e9/4b/243e-9c0c-44df-bd1f-6decc374cf78/data",
        "storage_class": "S",
        "size": 1562554,
        "checksum": "md5:3cc016be06f2be46d3a438db23c40bf3",
        "readable": True,
        "writable": False,
        "last_check_at": None,
        "last_check": True,
    }


@pytest.fixture(scope="function")
def fr_data(ov_data):
    """File record data."""
    return {
        "id": None,
        "json": {},
        "created": ov_data["created"],
        "updated": ov_data["updated"],
        "version_id": 1,
        "key": ov_data["key"],
        "record_id": None,
        "object_version_id": ov_data["version_id"],
    }



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Utils tests."""

import re

import pytest

from invenio_rdm_migrator.utils import dict_set, ts

###
# timestamp
###


def test_current_timestamp_iso():
    assert isinstance(ts(), str)


def test_current_timestamp():
    assert isinstance(ts(iso=False), float)


def test_current_timestamp_fmt():
    value = ts(fmt="%Y-%m-%dT%H:%M:%S")
    exp = re.compile(r"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}")
    assert exp.match(value)


###
# Dict set
###


def test_existing_root():
    source = {"a": {"b": 1}}
    key = "a.c"
    value = 2
    dict_set(source, key, value)
    assert source == {"a": {"b": 1, "c": 2}}


def test_set_existing_key():
    source = {"a": {"b": 1}}
    key = "a.b"
    value = 2
    dict_set(source, key, value)
    assert source == {"a": {"b": 2}}


def test_set_list():
    source = {"a": {}}
    key = "a.b"
    value = [1, 2]
    dict_set(source, key, value)
    assert source == {"a": {"b": [1, 2]}}


def test_set_empty_root():
    source = {}
    key = "a.b"
    value = 6
    dict_set(source, key, value)
    assert source == {"a": {"b": 6}}


def test_set_empty_root_list_key():
    source = {}
    key = ["a", "b"]
    value = 6
    dict_set(source, key, value)
    assert source == {"a": {"b": 6}}


def test_nested_dict():
    source = {"a": {"b": {}}}
    key = "a.b.c.d"
    value = 3
    dict_set(source, key, value)
    assert source == {"a": {"b": {"c": {"d": 3}}}}


def test_int_key():
    source = {1: ""}
    key = 1
    value = 3
    # "lookup must be string or list"
    pytest.raises(TypeError, dict_set, source, key, value)


def test_no_key():
    source = {"a": "value"}
    key = None
    value = 3
    # ("No lookup key specified"
    pytest.raises(KeyError, dict_set, source, key, value)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration user transform interfaces."""

from abc import abstractmethod

from ...transform import Entry, Transform


class UserTransform(Transform):
    """Base class for user data transformation."""

    @abstractmethod
    def _user(self, entry):
        """Transform the user."""
        pass

    @abstractmethod
    def _session_activity(self, entry):
        """Transform the session activity."""
        pass

    @abstractmethod
    def _tokens(self, entry):
        """Transform the tokens."""
        pass

    @abstractmethod
    def _applications(self, entry):
        """Transform the applications."""
        pass

    @abstractmethod
    def _oauth(self, entry):
        """Transform the OAuth accounts."""
        pass

    @abstractmethod
    def _identities(self, entry):
        """Transform the identities."""
        pass

    def _transform(self, entry):
        """Transform a single entry."""
        return {
            "user": self._user(entry),
            "session_activity": self._session_activity(entry),
            "tokens": self._tokens(entry),
            "applications": self._applications(entry),
            "oauth": self._oauth(entry),
            "identities": self._identities(entry),
        }


class UserEntry(Entry):
    """Transform a single user entry."""

    @abstractmethod
    def _id(self, entry):
        """Returns the user ID."""
        pass

    @abstractmethod
    def _created(self, entry):
        """Returns the creation date."""
        pass

    @abstractmethod
    def _updated(self, entry):
        """Returns the update date."""
        pass

    @abstractmethod
    def _version_id(self, entry):
        """Returns the version id.

        {Elastic,Open}search will fail with values lower than 1.
        """
        pass

    @abstractmethod
    def _email(self, entry):
        """Returns the email."""
        pass

    @abstractmethod
    def _active(self, entry):
        """Returns if the user is active."""
        pass

    @abstractmethod
    def _password(self, entry):
        """Returns the password."""
        pass

    @abstractmethod
    def _confirmed_at(self, entry):
        """Returns the confirmation date."""
        pass

    @abstractmethod
    def _blocked_at(self, entry):
        """Returns the blocking/inactivation date."""
        pass

    @abstractmethod
    def _verified_at(self, entry):
        """Returns the verification date."""
        pass

    @abstractmethod
    def _username(self, entry):
        """Returns the username."""
        pass

    @abstractmethod
    def _displayname(self, entry):
        """Returns the displayname."""
        pass

    @abstractmethod
    def _profile(self, entry):
        """Returns the profile."""
        pass

    @abstractmethod
    def _preferences(self, entry):
        """Returns the preferences."""
        pass

    @abstractmethod
    def _login_information(self, entry):
        """Returns the login information."""
        pass

    def transform(self, entry):
        """Transform a user single entry."""
        return {
            "id": self._id(entry),
            "created": self._created(entry),
            "updated": self._updated(entry),
            "version_id": self._version_id(entry),
            "email": self._email(entry),
            "active": self._active(entry),
            "password": self._password(entry),
            "confirmed_at": self._confirmed_at(entry),
            "blocked_at": self._blocked_at(entry),
            "verified_at": self._verified_at(entry),
            "username": self._username(entry),
            "displayname": self._displayname(entry),
            "profile": self._profile(entry),
            "preferences": self._preferences(entry),
            "login_information": self._login_information(entry),
        }



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""JSON field load module."""

import json


class JSONTransformMixin:
    """Loads json fields that are represented as strings."""

    @staticmethod
    def _load_json_fields(fields, data):
        """Load json fields."""
        for field in fields:
            data[field] = json.loads(data[field])



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration OAuth load module."""

from ...load.postgresql.bulk import PostgreSQLCopyLoad
from ...load.postgresql.bulk.generators import (
    ExistingDataTableGenerator,
    SingleTableGenerator,
)
from ..models.oauth import RemoteAccount, RemoteToken, ServerClient, ServerToken


class OAuthClientCopyLoad(PostgreSQLCopyLoad):
    """OAuthClient loading of existing data.

    It does not load the two tables with tokens, since they need rehashing.
    """

    def __init__(self, **kwargs):
        """Constructor."""
        super().__init__(
            table_generators=[
                # it passes on prepare so entries wont have an effect on it
                ExistingDataTableGenerator(tables=[RemoteAccount]),
                # this tg will treat the entries passing through the stream
                SingleTableGenerator(table=RemoteToken),
            ],
            **kwargs
        )


class OAuthServerCopyLoad(PostgreSQLCopyLoad):
    """OAuth2Server loading of existing data.

    It does not load the two tables with tokens, since they need rehashing.
    """

    def __init__(self, **kwargs):
        """Constructor."""
        super().__init__(
            table_generators=[
                # it passes on prepare so entries wont have an effect on it
                ExistingDataTableGenerator(tables=[ServerClient]),
                # this tg will treat the entries passing through the stream
                SingleTableGenerator(table=ServerToken),
            ],
            **kwargs
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Identifiers generators module."""

from uuid import uuid4

from ..state import STATE


def generate_uuid(data):
    """Generate a UUID."""
    return str(uuid4())


def pid_pk():
    """Generate an autoincrementing numeric primary key."""
    state = STATE.VALUES
    state_value = state.get("max_pid_pk")
    if not state_value:
        value = 1_000_000
        state.add("max_pid_pk", {"value": 1_000_000})
    else:
        value = state_value["value"] + 1
        state.update("max_pid_pk", {"value": value})

    return value


def generate_pk(data):
    """Generate a primary key."""
    return pid_pk()


def generate_pk_for(model_cls):
    """Generate an integer/serial primary key or a specific model.

    Generated PKs values are stored in the global values state under they key format
    "max_{model.__tablename__}_pk".
    """

    def _pk_gen(_):
        state = STATE.VALUES
        key = f"max_{model_cls.__tablename__}_pk"
        state_value = state.get(key)
        if not state_value:
            value = 1_000_000
            state.add(key, {"value": 1_000_000})
        else:
            value = state_value["value"] + 1
            state.update(key, {"value": value})
        return value

    return _pk_gen


def generate_recid(data, status="R"):
    """Generate a record id object."""
    # pk is not the pid_value, that comes from rec.json.id in the tg
    return {
        "pk": pid_pk(),
        # keep obj_type since the record needs this key dereferenced
        "obj_type": "rec",
        "pid_type": "recid",
        "status": status,
    }



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration transaction transform."""

from abc import ABC

from ..actions.base import TransformAction
from .base import Transform
from .errors import MultipleActionMatches, NoActionMatch


class BaseTxTransform(Transform, ABC):
    """Transform a transaction."""

    actions: list[TransformAction] = []

    def _detect_action(self, tx):
        match_cnt = 0
        match_cls = None
        for action_cls in self.actions:
            if action_cls.matches_action(tx):
                match_cnt += 1
                match_cls = action_cls

        if match_cnt > 1:
            raise MultipleActionMatches(tx)

        return match_cls  # return the first matched class

    def _transform(self, tx):
        """Transform action.

        Each action contains the raw transaction data.
        :returns: an instance of LoadAction.
        """
        action_cls = self._detect_action(tx)
        if not action_cls:
            raise NoActionMatch(tx)

        action = action_cls(tx)
        return action.transform()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Files models."""

from dataclasses import InitVar
from uuid import UUID

from sqlalchemy.orm import Mapped, mapped_column
from sqlalchemy.types import BigInteger

from ...load.postgresql.models import Model


class FilesBucket(Model):
    """Files bucket dataclass model."""

    __tablename__: InitVar[str] = "files_bucket"

    id: Mapped[UUID] = mapped_column(primary_key=True)
    created: Mapped[str]  # datetime
    updated: Mapped[str]  # datetime
    default_location: Mapped[int]
    default_storage_class: Mapped[str]
    size: Mapped[int]
    quota_size: Mapped[int] = mapped_column(BigInteger(), nullable=True)
    max_file_size: Mapped[int] = mapped_column(BigInteger(), nullable=True)
    locked: Mapped[bool]
    deleted: Mapped[bool]


class FilesInstance(Model):
    """Files instance dataclass model."""

    __tablename__: InitVar[str] = "files_files"

    id: Mapped[UUID] = mapped_column(primary_key=True)
    created: Mapped[str]  # datetime
    updated: Mapped[str]  # datetime
    uri: Mapped[str] = mapped_column(nullable=True)
    storage_class: Mapped[str] = mapped_column(nullable=True)
    size: Mapped[int] = mapped_column(nullable=True)
    checksum: Mapped[str] = mapped_column(nullable=True)
    readable: Mapped[bool]
    writable: Mapped[bool]
    last_check_at: Mapped[str] = mapped_column(nullable=True)  # datetime
    last_check: Mapped[bool]


class FilesObjectVersion(Model):
    """Files object version dataclass model."""

    __tablename__: InitVar[str] = "files_object"

    version_id: Mapped[UUID] = mapped_column(primary_key=True)
    created: Mapped[str]  # datetime
    updated: Mapped[str]  # datetime
    key: Mapped[str]
    bucket_id: Mapped[UUID]
    file_id: Mapped[UUID] = mapped_column(nullable=True)
    _mimetype: Mapped[str] = mapped_column(nullable=True)
    is_head: Mapped[bool]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration extract interfaces."""


from abc import ABC, abstractmethod


class Extract(ABC):
    """Base class for data extraction."""

    @abstractmethod
    def run(self):  # pragma: no cover
        """Yield one element at a time."""
        pass



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""File actions tests."""

from uuid import UUID

import pytest

from invenio_rdm_migrator.load.postgresql.transactions.operations import OperationType
from invenio_rdm_migrator.streams.actions.load import DraftFileUploadAction
from invenio_rdm_migrator.streams.models.files import (
    FilesBucket,
    FilesInstance,
    FilesObjectVersion,
)
from invenio_rdm_migrator.streams.models.records import RDMDraftFile


def test_upload_file_action(
    buckets_state, bucket_data_w_size, fi_data, ov_data, fr_data
):
    data = dict(
        tx_id=1,
        bucket=bucket_data_w_size,
        object_version=ov_data,
        file_instance=fi_data,
        file_record=fr_data,
    )
    action = DraftFileUploadAction(data)
    rows = list(action.prepare())

    assert len(rows) == 4
    assert rows[0].type == OperationType.UPDATE
    assert rows[0].model == FilesBucket
    assert rows[1].type == OperationType.INSERT
    assert rows[1].model == FilesObjectVersion
    assert rows[2].type == OperationType.INSERT
    assert rows[2].model == FilesInstance
    assert rows[3].type == OperationType.INSERT
    assert rows[3].model == RDMDraftFile
    UUID(rows[3].data["id"])  # would raise value error if not UUID or None
    # still type str since it has not been inserted in DB
    assert rows[3].data["record_id"] == "d94f793c-47d2-48e2-9867-ca597b4ebb41"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.
"""Community stream fixtures."""

from copy import deepcopy

import pytest


@pytest.fixture(scope="module")
def transformed_community_entry():
    """Community data as after being transformed."""
    return {
        "community": {
            "created": "2023-01-01 12:00:00.00000",
            "updated": "2023-01-31 12:00:00.00000",
            "slug": "migrator",
            "version_id": 1,
            "json": {
                "title": "Migrator community",
                "description": "Migrator testing community",
                "page": "",
                "curation_policy": "",
            },
            "bucket_id": "12345678-abcd-1a2b-3c4d-123abc456def",
        },
        "oai_set": {
            "created": "2023-01-01 12:00:00.00000",
            "updated": "2023-01-31 12:00:00.00000",
            "id": 1,
            "spec": "community-migrator",
            "name": "Migrator community",
            "description": "Migrator testing community",
        },
        "community_members": [
            {
                "created": "2023-01-01 12:00:00.00000",
                "updated": "2023-01-31 12:00:00.00000",
                "json": {},
                "version_id": 1,
                "role": "owner",
                "visible": True,
                "active": True,
                "user_id": 1,
                "group_id": None,
                "request_id": None,
            }
        ],
        "featured_community": {},
        "community_files": {
            "file": {
                "created": "2023-01-01 12:00:00.00000",
                "updated": "2023-01-31 12:00:00.00000",
                "id": "12345678-abcd-1a2b-3c4d-123abc456def",
                "json": {},
                "version_id": 1,
                "key": "logo",
            },
            "bucket": {
                "created": "2023-01-01 12:00:00.00000",
                "updated": "2023-01-31 12:00:00.00000",
                "default_location": None,
                "default_storage_class": "L",
                "size": 0,
                "quota_size": 0,
                "max_file_size": 0,
                "locked": False,
                "deleted": False,
                "id": "12345678-abcd-1a2b-3c4d-123abc456def",
            },
            "file_object": {
                "created": "2023-01-01 12:00:00.00000",
                "updated": "2023-01-31 12:00:00.00000",
                "bucket_id": "12345678-abcd-1a2b-3c4d-123abc456def",
                "version_id": "12345678-abcd-1a2b-3c4d-123abc456def",
                "key": "logo",
                "file_id": "12345678-abcd-1a2b-3c4d-123abc456def",
                "_mimetype": None,
                "is_head": True,
            },
        },
    }


@pytest.fixture(scope="module")
def transformed_community_entry_pks(transformed_community_entry):
    """Request data as after being transformed and passed by generate_pks.

    The last step would happen in the table_generator._prepare.
    """
    data = deepcopy(transformed_community_entry)
    data["community"]["id"] = "7357c033-abcd-1a2b-3c4d-123abc456def"
    data["community_members"][0]["id"] = "12345678-abcd-1a2b-3c4d-123abc456def"

    return data



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Base state module."""

from abc import ABC
from pathlib import Path
from uuid import UUID

import sqlalchemy as sa
from sqlalchemy_utils.types import JSONType, UUIDType

from .logging import Logger
from .utils import ts


class StateDB:
    """Migration state."""

    def __init__(self, db_dir, validators=None):
        """Constructor.

        :param db_dir: path to sqlite db file.
        """
        # for extra validation that cannot be achieved with SQL constraints
        self.validators = validators or {}
        self.db_dir = Path(db_dir)
        self.db_filepath = self.db_dir / "state.db"
        self._mem_eng = None
        # metadata cannot be initialized lazily since tables is access before mem_eng
        # in any op (get, all, add, etc.)
        self._metadata = sa.MetaData()
        self._initialize_db(self._metadata)  # bind tables to metadata

    @property
    def mem_eng(self):
        """Singleton property for memory engine.

        Lazy loading of data from disk.
        """
        if not self._mem_eng:
            self._mem_eng = sa.create_engine("sqlite:///:memory:")

            self._load_from_disk()

        return self._mem_eng

    @property
    def tables(self):
        """Return the currently registered tables in the metadata."""
        return self._metadata.tables

    def _load_from_disk(self):
        """Loads the database from disk into memory."""
        start = ts(iso=False)
        logger = Logger.get_logger()
        logger.info(f"Loading state from {self.db_filepath}.")

        disk_eng = sa.create_engine(f"sqlite:///{self.db_filepath}")
        self._metadata.create_all(disk_eng)  # create tables if not exist

        self._copy_db(disk_eng, self.mem_eng)  # dump state into in-memory sqlite db

        end = ts(iso=False)
        logger.info("Finished loading state.")
        logger.info(f"State loading took {end-start} seconds.")

    def save(self, filename=None, filepath=None):
        """Save the current in-memory state to disk.

        It will overwrite the loaded db file.
        :param filename: name of the state file, it will be save to db_dir.
        :param filepath: full path, including file name, where to save the state db.
        """
        logger = Logger.get_logger()
        start = ts(iso=False)

        # move current db state to backup
        disk_file = self.db_filepath
        if filepath:  # support overwriting the path
            disk_file = Path(filepath)
        elif filename:  # support overwriting the file name
            disk_file = self.db_dir / filename

        logger.info(f"Dumping state to {disk_file}.")
        backup_file = Path(self.db_filepath.name + ".backup")
        if self.db_filepath.exists():
            self.db_filepath.rename(backup_file)

        # dump state into from sqlite db
        disk_eng = sa.create_engine(f"sqlite:///{disk_file}")
        self._copy_db(self.mem_eng, disk_eng)

        # delete old db (backup)
        backup_file.unlink(missing_ok=True)

        end = ts(iso=False)
        logger.info("Finished dumping state.")
        logger.info(f"State dumping took {end-start} seconds.")

    def get(self, table_name, key_attr, key_value):
        """Query a table by key."""
        table = self.tables[table_name]
        with self.mem_eng.connect() as conn:
            result = conn.execute(
                sa.select(table).where(getattr(table.columns, key_attr) == key_value)
            ).one_or_none()
        return result

    def search(self, table_name, column, value):
        """Query a table by column."""
        table = self.tables[table_name]
        with self.mem_eng.connect() as conn:
            result = conn.execute(
                sa.select(table).where(getattr(table.columns, column) == value)
            ).all()

        return result

    def all(self, table_name):
        """Get all the data from the state.

        :returns: a cursor result.
        """
        table = self.tables[table_name]

        with self.mem_eng.connect() as conn:
            result = conn.execute(sa.select(table))
        return result

    def add(self, table_name, data):
        """Add key,data pair to the state."""
        table = self.tables[table_name]
        with self.mem_eng.connect() as conn:
            result = conn.execute(sa.insert(table).values(**data))
            conn.commit()

        assert result.rowcount == 1

    def delete(self, table_name, key_attr, key_value):
        """Delete an item from the state."""
        table = self.tables[table_name]

        with self.mem_eng.connect() as conn:
            result = conn.execute(
                sa.delete(table).where(getattr(table.columns, key_attr) == key_value)
            )
            conn.commit()

        assert result.rowcount == 1  # the update succeeded

    def update(self, table_name, key_attr, key_value, data):
        """Update an entry."""
        table = self.tables[table_name]

        with self.mem_eng.connect() as conn:
            result = conn.execute(
                sa.update(table)
                .where(getattr(table.columns, key_attr) == key_value)
                .values(**data)
            )
            conn.commit()

        assert result.rowcount == 1  # the update succeeded

    def _copy_db(cls, src, dest):
        """Copy source database into destination."""
        # sqlalchemy abstraction does not have a backup we need a raw connection
        src_conn = src.raw_connection()
        dst_conn = dest.raw_connection()

        src_conn.backup(dst_conn.connection)

        src_conn.close()
        dst_conn.close()

    def validate(self, table_name, data):
        """Validate data for a specific table."""
        validator = self.validators.get(table_name)
        if validator:
            validator.validate(data)

    @classmethod
    def _initialize_db(cls, metadata):
        """Initialize state tables and bind to metadata object."""
        # this registration method is not ideal. It implies a tight coupling between
        # state and which streams can be run, and there is no independence (i.e. running one
        # stream means having all state tables defined rather than only those needed).
        # The tables should be defined per stream (e.g. streams/records/state.py:ParentTable)
        # and then be injected into the state.

        sa.Table(
            "parents",
            metadata,
            sa.Column("recid", sa.String, primary_key=True),
            sa.Column("id", UUIDType, nullable=False, unique=True),
            sa.Column("latest_id", UUIDType),
            # could use default=0, but it could create uncontrolled scenarios
            sa.Column("latest_index", sa.Integer),
            sa.Column("next_draft_id", UUIDType),
            sqlite_autoincrement=False,
        )

        sa.Table(
            "records",
            metadata,
            sa.Column("recid", sa.String, primary_key=True),
            sa.Column("id", UUIDType, nullable=False, unique=True),
            sa.Column("parent_id", UUIDType, nullable=False),
            sa.Column("index", sa.Integer, nullable=False),
            # record's version_id
            sa.Column("fork_version_id", sa.Integer, nullable=False),
            sa.Column("pids", JSONType, default="{}"),
            sqlite_autoincrement=False,
        )

        sa.Table(
            "buckets",
            metadata,
            sa.Column("id", UUIDType, primary_key=True),
            sa.Column("draft_id", UUIDType, nullable=False),
            sqlite_autoincrement=False,
        )

        sa.Table(
            "file_records",
            metadata,
            sa.Column("id", UUIDType, primary_key=True),
            sa.Column("json", JSONType, default="{}"),
            sa.Column("created", sa.String, nullable=False),
            sa.Column("updated", sa.String, nullable=False),
            sa.Column("version_id", sa.Integer, nullable=False),
            sa.Column("key", sa.String, nullable=False),
            sa.Column("record_id", UUIDType, nullable=False),
            sa.Column("object_version_id", UUIDType, nullable=True),
            sqlite_autoincrement=False,
        )

        sa.Table(
            "communities",
            metadata,
            sa.Column("slug", sa.String, primary_key=True),
            sa.Column("id", UUIDType, unique=True, nullable=False),
            sa.Column("bucket_id", UUIDType, unique=True, nullable=False),
            sa.Column("oai_set_id", sa.Integer, unique=True, nullable=False),
            sa.Column("community_file_id", UUIDType, unique=True),
            sa.Column("logo_object_version_id", UUIDType, unique=True),
            sqlite_autoincrement=False,
        )

        sa.Table(
            "pids",
            metadata,
            sa.Column("pid_value", sa.String, primary_key=True),
            sa.Column("id", sa.Integer, unique=True, nullable=False),
            sa.Column("pid_type", sa.String, nullable=False),
            sa.Column("status", sa.String, nullable=False),
            sa.Column("created", sa.String, nullable=False),
            # keep obj_type since the record needs this key dereferenced
            sa.Column("obj_type", sa.String),
            sqlite_autoincrement=False,
        )

        # e.g. key = max_pid_pk, value = 1_000_000
        sa.Table(
            "global",
            metadata,
            sa.Column("key", sa.String, primary_key=True),
            sa.Column("value", sa.Integer),
            sqlite_autoincrement=False,
        )


class StateEntity:
    """Helper class to query specific state tables."""

    def __init__(self, state, table_name, pk_attr, cache=True):
        """Constructor."""
        self.state = state
        self.pk_attr = pk_attr
        self.table_name = table_name
        self._cache = None
        if cache:
            self._cache = {}
            self._init_cache()

    def _init_cache(self):
        """Initialize the cache from state."""
        for row in self.state.all(self.table_name):
            data = self._row_as_dict(row)
            key = data[self.pk_attr]
            self._cache[key] = data

    @classmethod
    def _row_as_dict(cls, row):
        """Transform a Row or LegacyRow into a dict."""
        # needed for backwards compatibility
        row_dict = {}
        if not row:
            return row_dict

        for key, value in row._mapping.items():
            # a better generalization might be wanted, for now we only have
            # string, int, JSON and UUID as column types. the only one cause troubles
            # is uuid, which needs to be stringified.
            if value and isinstance(value, UUID):
                value = str(value)
            row_dict[key] = value

        return row_dict

    def get(self, key):
        """Get a row by key."""
        if self._cache is not None:
            return self._cache.get(key, {})
        return self._row_as_dict(self.state.get(self.table_name, self.pk_attr, key))

    def search(self, column, value):
        """Search rows."""
        for row in self.state.search(self.table_name, column, value):
            yield self._row_as_dict(row)

    def all(self):
        """Get all rows."""
        if self._cache is not None:
            for v in self._cache.values():
                yield v
        else:
            for row in self.state.all(self.table_name):
                yield self._row_as_dict(row)

    def add(self, key, data):
        """Add data row."""
        self.state.validate(self.table_name, data)
        if self._cache is not None:
            if key in self._cache:
                raise ValueError("Key {key} already in state.")
            self._cache[key] = data
        else:
            self.state.add(self.table_name, {self.pk_attr: key, **data})

    def update(self, key, data):
        """Add data row."""
        self.state.validate(self.table_name, data)
        if self._cache is not None:
            self._cache[key].update(data)
        else:
            self.state.update(
                self.table_name, self.pk_attr, key, {self.pk_attr: key, **data}
            )

    def delete(self, key):
        """Delete data row."""
        if self._cache is not None:
            self._cache.pop(key, None)
        else:
            self.state.delete(self.table_name, self.pk_attr, key)


class StateValidator(ABC):
    """Validator interface."""

    @classmethod
    def validate(cls, data):
        """Data validation."""
        # abc fails on instantiation, e.g. Test(), since this is a classmethod
        # it will always call the parent, therefore the default is False.
        return False


class STATE:
    """Static class holding a reference to a state query."""

    # implemented this way to avoid having to drill state.global from the
    # runner to the lower level pid_pk()
    # ideally there would be some sort of application context/proxies to access this.
    VALUES = None
    PARENTS = None
    RECORDS = None
    COMMUNITIES = None
    PIDS = None
    BUCKETS = None
    FILE_RECORDS = None

    @classmethod
    def initialized_state(cls, state_db, cache=True):
        """Initializes state."""
        cls.PARENTS = StateEntity(state_db, "parents", "recid", cache=cache)
        cls.RECORDS = StateEntity(state_db, "records", "recid", cache=cache)
        cls.BUCKETS = StateEntity(state_db, "buckets", "id", cache=cache)
        cls.FILE_RECORDS = StateEntity(state_db, "file_records", "id", cache=cache)
        cls.COMMUNITIES = StateEntity(state_db, "communities", "slug", cache=cache)
        cls.PIDS = StateEntity(state_db, "pids", "pid_value", cache=cache)
        cls.VALUES = StateEntity(state_db, "global", "key", cache=cache)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration load actions."""

from .communities import (
    CommunityCreateAction,
    CommunityDeleteAction,
    CommunityUpdateAction,
)
from .drafts import DraftCreateAction, DraftEditAction, DraftPublishAction
from .files import DraftFileUploadAction
from .users import UserDeactivationAction, UserEditAction, UserRegistrationAction

__all__ = (
    "CommunityCreateAction",
    "CommunityDeleteAction",
    "CommunityUpdateAction",
    "DraftCreateAction",
    "DraftEditAction",
    "DraftFileUploadAction",
    "DraftPublishAction",
    "UserDeactivationAction",
    "UserEditAction",
    "UserRegistrationAction",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Requests load tests."""

import os
from unittest.mock import patch

import pytest

from invenio_rdm_migrator.streams.requests import RequestCopyLoad


@pytest.fixture(scope="function")
def request_copy_load(communities_state, tmp_dir):
    """Request load instance."""
    # the db queries will be mocked
    load = RequestCopyLoad(db_uri="None", tmp_dir=tmp_dir.name)
    yield load
    load._cleanup()


@patch(
    "invenio_rdm_migrator.load.ids.uuid4",
    lambda: "12345678-abcd-1a2b-3c4d-123abc456def",
)
def test_request_load_prepare(request_copy_load, transformed_incl_req_entry):
    """Test the table preparation (file creation)."""
    tables = [
        table for _, table in request_copy_load._prepare([transformed_incl_req_entry])
    ]

    # assert tables
    assert len(tables) == 1
    assert tables[0].__tablename__ == "request_metadata"

    # assert files were created and have the content
    files = list(os.scandir(request_copy_load.tmp_dir))
    assert len(files) == 1

    # assert request metadata content
    with open(f"{request_copy_load.tmp_dir}/request_metadata.csv", "r") as file:
        # uuid, json, created, updated, version_id, number, expired_at
        expected = (
            "12345678-abcd-1a2b-3c4d-123abc456def,"
            '"{""type"":""community-inclusion"",""title"":""title"",""topic"":{""record"":""123456""},""status"":""submitted"",""receiver"":{""community"":""12345678-abcd-1a2b-3c4d-123abc456def""},""created_by"":{""user"":""3""},""$schema"":""local://requests/request-v1.0.0.json""}"'
            ",01/01/2023,01/01/2023,1,1,01/01/2023\n"
        )
        content = file.read()
        assert content == expected



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Base transformation tests."""

from types import GeneratorType

import pytest

from invenio_rdm_migrator.logging import Logger
from invenio_rdm_migrator.transform import Transform
from invenio_rdm_migrator.transform.base import drop_nones

###
# Base class
###


def test_cannot_instantiate_base_class():
    with pytest.raises(TypeError):
        Transform()

    class Test(Transform):
        """Test Transform."""

        pass

    with pytest.raises(TypeError):
        Test()


class TestTransform(Transform):
    def _transform(self, entry):
        return entry * 2


def test_transform_returns_transformed_entry():
    t = TestTransform()
    assert list(t.run([1])) == [2]


def test_run_returns_iterator():
    extract = TestTransform()
    assert isinstance(extract.run([]), GeneratorType)


def test_run_continues_to_next_entry_when_exception_raised(capsys, tmp_path):
    Logger.initialize(tmp_path)

    class TestTransform(Transform):
        def _transform(self, entry):
            if entry == 2:
                raise Exception("Test exception")
            return entry

    t = TestTransform()
    # should raise and log exception on 2, log it and continue to 3
    assert list(t.run([1, 2, 3])) == [1, 3]
    captured = capsys.readouterr()
    assert "Test exception" in captured.err


###
# drop_nones
###


def test_simple_dict_no_nones():
    data = {"a": 1, "b": "hello", "c": [1, 2, 3]}
    assert drop_nones(data) == data


def test_nested_dict_no_nones():
    data = {"a": 1, "b": {"c": 2, "d": {"e": "hello"}}}
    assert drop_nones(data) == data


def test_empty_dict():
    data = {}
    assert drop_nones(data) == {}


def test_nested_dict_first_level_nones():
    data = {"a": 1, "b": {"c": None, "d": {"e": "hello", "f": None}}, "g": None}
    expected = {
        "a": 1,
        "b": {"d": {"e": "hello"}},
    }
    assert drop_nones(data) == expected


def test_nested_dict_some_nones():
    data = {"a": 1, "b": {"c": None, "d": {"e": "hello", "f": None}}, "g": None}
    expected = {
        "a": 1,
        "b": {"d": {"e": "hello"}},
    }
    assert drop_nones(data) == expected


def test_dict_only_nones():
    data = {"a": None, "b": None, "c": None}
    assert drop_nones(data) == {}



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration record stream."""

from .load import RDMRecordCopyLoad
from .transform import RDMRecordEntry, RDMRecordTransform

__all__ = (
    "RDMRecordCopyLoad",
    "RDMRecordEntry",
    "RDMRecordTransform",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Dataclasses record models to generate table rows."""

from dataclasses import InitVar
from uuid import UUID

from sqlalchemy.orm import Mapped, mapped_column

from ...load.postgresql.models import Model


class RDMRecordMetadata(Model):
    """RDM Record Metadata dataclass model."""

    __tablename__: InitVar[str] = "rdm_records_metadata"

    id: Mapped[UUID] = mapped_column(primary_key=True)
    json: Mapped[dict] = mapped_column(nullable=True)
    created: Mapped[str]  # datetime
    updated: Mapped[str]  # datetime
    version_id: Mapped[int]
    index: Mapped[int]
    bucket_id: Mapped[UUID]
    media_bucket_id: Mapped[UUID] = mapped_column(nullable=True)
    parent_id: Mapped[UUID]
    deletion_status: Mapped[str]


class RDMParentMetadata(Model):
    """RDM Parent Metadata dataclass model."""

    __tablename__: InitVar[str] = "rdm_parents_metadata"

    id: Mapped[UUID] = mapped_column(primary_key=True)
    json: Mapped[dict] = mapped_column(nullable=True)
    created: Mapped[str]  # datetime
    updated: Mapped[str]  # datetime
    version_id: Mapped[int]


class RDMVersionState(Model):
    """RDM Version State dataclass model."""

    __tablename__: InitVar[str] = "rdm_versions_state"

    parent_id: Mapped[UUID] = mapped_column(primary_key=True)
    latest_id: Mapped[UUID] = mapped_column(nullable=True)
    next_draft_id: Mapped[UUID] = mapped_column(nullable=True)
    # latest_index has a programmatic default of 1 if the db value is None
    latest_index: Mapped[int] = mapped_column(nullable=True)


class RDMDraftMetadata(Model):
    """RDM Draft Metadata dataclass model."""

    __tablename__: InitVar[str] = "rdm_drafts_metadata"

    id: Mapped[UUID] = mapped_column(primary_key=True)
    json: Mapped[dict] = mapped_column(nullable=True)
    created: Mapped[str]  # datetime
    updated: Mapped[str]  # datetime
    version_id: Mapped[int]
    index: Mapped[int]
    bucket_id: Mapped[UUID]
    media_bucket_id: Mapped[UUID] = mapped_column(nullable=True)
    parent_id: Mapped[UUID]
    expires_at: Mapped[str]
    # in a new version this value is None
    fork_version_id: Mapped[int] = mapped_column(nullable=True)


class RDMRecordFile(Model):
    """RDM Record File dataclass model."""

    __tablename__: InitVar[str] = "rdm_records_files"

    id: Mapped[UUID] = mapped_column(primary_key=True)
    json: Mapped[dict]
    created: Mapped[str]  # datetime
    updated: Mapped[str]  # datetime
    version_id: Mapped[int]
    key: Mapped[str]
    record_id: Mapped[UUID]
    object_version_id: Mapped[UUID]


class RDMRecordMediaFile(Model):
    """RDM Record Media File dataclass model."""

    __tablename__: InitVar[str] = "rdm_records_media_files"

    id: Mapped[UUID] = mapped_column(primary_key=True)
    json: Mapped[dict] = mapped_column(nullable=True)
    created: Mapped[str]  # datetime
    updated: Mapped[str]  # datetime
    version_id: Mapped[int]
    key: Mapped[str]
    record_id: Mapped[UUID]
    object_version_id: Mapped[UUID]


class RDMDraftFile(Model):
    """RDM Draft File dataclass model."""

    __tablename__: InitVar[str] = "rdm_drafts_files"

    # duplicated code to avoid dealing with sqlalchemy inheritance and fks
    # which in our case do not play any role
    id: Mapped[UUID] = mapped_column(primary_key=True)
    json: Mapped[dict] = mapped_column(nullable=True)
    created: Mapped[str]  # datetime
    updated: Mapped[str]  # datetime
    version_id: Mapped[int]
    key: Mapped[str]
    record_id: Mapped[UUID]
    object_version_id: Mapped[UUID] = mapped_column(nullable=True)


class RDMDraftMediaFile(Model):
    """RDM Draft Media File dataclass model."""

    __tablename__: InitVar[str] = "rdm_drafts_media_files"

    id: Mapped[UUID] = mapped_column(primary_key=True)
    json: Mapped[dict] = mapped_column(nullable=True)
    created: Mapped[str]  # datetime
    updated: Mapped[str]  # datetime
    version_id: Mapped[int]
    key: Mapped[str]
    record_id: Mapped[UUID]
    object_version_id: Mapped[UUID] = mapped_column(nullable=True)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration OAuth stream."""

from .load import OAuthClientCopyLoad, OAuthServerCopyLoad
from .transform import OAuthRemoteTokenTransform, OAuthServerTokenTransform

__all__ = (
    "OAuthClientCopyLoad",
    "OAuthRemoteTokenTransform",
    "OAuthServerCopyLoad",
    "OAuthServerTokenTransform",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Dataclasses user models to generate table rows."""

from dataclasses import InitVar

from sqlalchemy.orm import Mapped, mapped_column

from ...load.postgresql.models import Model


class User(Model):
    """User dataclass model."""

    __tablename__: InitVar[str] = "accounts_user"

    id: Mapped[int] = mapped_column(primary_key=True)
    created: Mapped[str]  # datetime
    updated: Mapped[str]  # datetime
    username: Mapped[str] = mapped_column(nullable=True)
    displayname: Mapped[str] = mapped_column(nullable=True)
    email: Mapped[str]
    password: Mapped[str]
    active: Mapped[bool]
    confirmed_at: Mapped[str] = mapped_column(nullable=True)  # datetime
    blocked_at: Mapped[str] = mapped_column(nullable=True)  # datetime
    verified_at: Mapped[str] = mapped_column(nullable=True)  # datetime
    version_id: Mapped[int]
    profile: Mapped[dict] = mapped_column(nullable=True)
    preferences: Mapped[dict] = mapped_column(nullable=True)


class LoginInformation(Model):
    """Login information dataclass model."""

    __tablename__: InitVar[str] = "accounts_user_login_information"

    user_id: Mapped[int] = mapped_column(primary_key=True)
    # the initial creation has them all to None
    last_login_at: Mapped[str] = mapped_column(nullable=True)  # datetime
    current_login_at: Mapped[str] = mapped_column(nullable=True)  # datetime
    last_login_ip: Mapped[str] = mapped_column(nullable=True)
    current_login_ip: Mapped[str] = mapped_column(nullable=True)
    login_count: Mapped[int]


class SessionActivity(Model):
    """Session activity dataclass model."""

    __tablename__: InitVar[str] = "accounts_user_session_activity"

    user_id: Mapped[int]
    created: Mapped[str]  # datetime
    updated: Mapped[str]  # datetime
    sid_s: Mapped[str] = mapped_column(primary_key=True)
    ip: Mapped[str] = mapped_column(nullable=True)
    country: Mapped[str] = mapped_column(nullable=True)
    browser: Mapped[str] = mapped_column(nullable=True)
    browser_version: Mapped[str] = mapped_column(nullable=True)
    os: Mapped[str] = mapped_column(nullable=True)
    device: Mapped[str] = mapped_column(nullable=True)


class UserIdentity(Model):
    """User identity dataclass model."""

    __tablename__: InitVar[str] = "accounts_useridentity"

    created: Mapped[str]  # datetime
    updated: Mapped[str]  # datetime
    id: Mapped[str] = mapped_column(primary_key=True)
    method: Mapped[str] = mapped_column(primary_key=True)
    id_user: Mapped[int]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration record table load module."""

from datetime import datetime
from uuid import UUID

import psycopg

from ....load.ids import generate_recid, generate_uuid, pid_pk
from ....load.postgresql.bulk.generators import TableGenerator
from ....logging import Logger
from ....state import STATE
from ...models.communities import RDMParentCommunityMetadata
from ...models.files import FilesObjectVersion
from ...models.pids import PersistentIdentifier
from ...models.records import (
    RDMParentMetadata,
    RDMRecordFile,
    RDMRecordMediaFile,
    RDMRecordMetadata,
)
from .parents import generate_parent_rows
from .references import CommunitiesReferencesMixin


def _is_valid_uuid(value):
    try:
        UUID(value)
        return True
    except Exception:
        return False


def generate_files_uuids(data):
    """Generate uuid for every file in the list.

    Return the transformed list with the generated ids.
    """
    _files = data["record_files"]
    for _file in _files:
        _file["id"] = generate_uuid({})
    return _files


def generate_record_uuid(data):
    """Generate record uuid if not present."""
    _id = data["record"].get("id")
    return _id if _id else generate_uuid(None)


class RDMRecordTableGenerator(TableGenerator, CommunitiesReferencesMixin):
    """RDM Record and related tables load."""

    def __init__(self):
        """Constructor."""
        super().__init__(
            tables=[
                PersistentIdentifier,
                RDMParentMetadata,
                RDMRecordMetadata,
                RDMParentCommunityMetadata,
                RDMRecordFile,
            ],
            post_load_hooks=[self.insert_media_files],
            pks=[
                ("record.id", generate_record_uuid),
                ("parent.id", generate_uuid),
                ("record.json.pid", generate_recid),
                ("parent.json.pid", generate_recid),
                ("record.parent_id", lambda d: d["parent"]["id"]),
                ("record_files", generate_files_uuids),
            ],
        )

    def _generate_rows(self, data, **kwargs):
        """Generates rows for a record."""
        now = datetime.utcnow().isoformat()
        parent = data["parent"]
        record = data.get("record")

        if not record:
            return

        record_pid = record["json"]["pid"]

        # parent
        state_parent = STATE.PARENTS.get(parent["json"]["id"])
        if not state_parent:
            STATE.PARENTS.add(
                parent["json"]["id"],  # recid
                {
                    "id": parent["id"],
                    "latest_index": record["index"],
                    "latest_id": record["id"],
                },
            )
            yield from generate_parent_rows(parent)
            # parent community
            if "default" in parent["json"]["communities"]:
                parent_def_id = parent["json"]["communities"]["default"]
                parent_comm_id = parent["id"]
                # when a community is deleted, its id is the slug and fails on DB
                # FK expects uuid and not string
                if _is_valid_uuid(parent_def_id) and _is_valid_uuid(parent_comm_id):
                    yield RDMParentCommunityMetadata(
                        community_id=parent_def_id,
                        record_id=parent_comm_id,
                        request_id=None,
                    )
                else:
                    record_id = record["json"]["id"]
                    logger = Logger.get_logger()
                    logger.error(
                        f"Record parent community not migrated. [{record_id=}]. "
                        f"[{parent_comm_id=}], default community [{parent_def_id}]"
                    )
        else:
            if state_parent.get("latest_index") < record["index"]:
                STATE.PARENTS.update(
                    parent["json"]["id"],
                    {
                        "latest_index": record["index"],
                        "latest_id": record["id"],
                    },
                )

        parent_id = state_parent["id"] if state_parent else record["parent_id"]
        # record
        STATE.RECORDS.add(
            record["json"]["id"],  # recid
            {
                "index": record["index"],
                "id": record["id"],  # uuid
                "parent_id": parent_id,  # parent uuid
                "fork_version_id": record["version_id"],
                "pids": record["json"]["pids"],
            },
        )

        yield RDMRecordMetadata(
            id=record["id"],
            json=record["json"],
            created=record["created"],
            updated=record["updated"],
            version_id=record["version_id"],
            index=record["index"],
            bucket_id=record["bucket_id"],
            media_bucket_id=record.get("media_bucket_id"),
            parent_id=parent_id,
            deletion_status="P",
        )
        # recid
        yield PersistentIdentifier(
            id=record_pid["pk"],
            pid_type=record_pid["pid_type"],
            pid_value=record["json"]["id"],
            status=record_pid["status"],
            object_type=record_pid["obj_type"],
            object_uuid=record["id"],
            created=now,
            updated=now,
        )
        # DOI
        if "doi" in record["json"]["pids"]:
            yield PersistentIdentifier(
                id=pid_pk(),
                pid_type="doi",
                pid_value=record["json"]["pids"]["doi"]["identifier"],
                status="R",
                object_type="rec",
                object_uuid=record["id"],
                created=now,
                updated=now,
            )
        # OAI
        if "oai" in record["json"]["pids"]:
            yield PersistentIdentifier(
                id=pid_pk(),
                pid_type="oai",
                pid_value=record["json"]["pids"]["oai"]["identifier"],
                status="R",
                object_type="rec",
                object_uuid=record["id"],
                created=now,
                updated=now,
            )

        # record files
        record_files = data["record_files"]
        for _file in record_files:
            yield RDMRecordFile(
                id=_file["id"],
                json=_file["json"],
                created=_file["created"],
                updated=_file["updated"],
                version_id=_file["version_id"],
                key=_file["key"],
                record_id=record["id"],
                object_version_id=_file["object_version_id"],
            )

    def _resolve_references(self, data, **kwargs):
        """Resolve references e.g communities slug names."""
        if "record" in data:
            # resolve parent communities slug
            parent = data["parent"]
            communities = parent["json"].get("communities")
            if communities:
                self.resolve_communities(communities)

    def insert_media_files(self, db_uri=None):
        """Inserts record media files from buckets and object version."""
        assert db_uri  # should have come from kwargs

        with psycopg.connect(db_uri) as conn:
            # the query needs to be split in 3 parts because the empty jsonb dict
            # would cause problems with the string formatting
            insert = f"""
                INSERT INTO {RDMRecordMediaFile.__tablename__} (
                    id, json, created, updated, version_id, key, record_id, object_version_id
                )
            """
            select = "SELECT gen_random_uuid(), '{}'::jsonb, rrm.created, rrm.updated, 1, fo.key, rrm.id, fo.version_id"
            from_and_join = f"""
                FROM {RDMRecordMetadata.__tablename__} AS rrm
                INNER JOIN {FilesObjectVersion.__tablename__} AS fo
                ON rrm.media_bucket_id = fo.bucket_id AND fo.is_head = 'true'
                WHERE rrm.media_bucket_id IS NOT NULL
            """
            # no return check, will raise if the sql statement fails
            # id and json do not have defaults in DB even though if the programmatic
            # models have them, so they need to be calculated
            conn.execute(insert + select + from_and_join)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""JSONLines extract tests."""

from pathlib import Path
from types import GeneratorType

import jsonlines
import pytest

from invenio_rdm_migrator.extract.jsonlines import JSONLExtract


@pytest.fixture(scope="function")
def jsonlines_file(tmp_dir):
    """Returns the file path of a jsonlines file."""
    filename = Path(tmp_dir.name) / "test.jsonl"
    with jsonlines.open(filename, mode="w") as writer:
        writer.write({"key": "1", "value": "one"})
        writer.write({"key": "2", "value": "two"})

    # note there is no need for cleanup since tmp will do so and it's function scoped
    yield filename


def test_constructor_sets_filepath_attribute(jsonlines_file):
    extract = JSONLExtract(jsonlines_file)
    assert extract.filepath == jsonlines_file


def test_run_returns_iterator(jsonlines_file):
    extract = JSONLExtract(filepath=jsonlines_file)
    assert isinstance(extract.run(), GeneratorType)


def test_constructor_raises_FileNotFoundError():
    with pytest.raises(FileNotFoundError):
        JSONLExtract("invalid_filepath.jsonl")


def test_reads_jsonlines_file(jsonlines_file):
    extract = JSONLExtract(filepath=jsonlines_file)
    assert len(list(extract.run())) == 2



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration load interfaces."""


from abc import ABC, abstractmethod


class Load(ABC):
    """Base class for data loading."""

    def _prepare(self, *args, **kwargs):  # pragma: no cover
        """Prepare data for loading."""
        pass

    @abstractmethod
    def _load(self, *args, **kwargs):  # pragma: no cover
        """load data."""
        pass

    @abstractmethod
    def _cleanup(self):  # pragma: no cover
        """Cleanup data after loading."""
        pass

    def _validate(self, *args, **kwargs):
        """Validate data before loading."""
        # It should default to False or be abstract to force implementation.
        # However, being pragmatic and due to time/resources constraints this validation
        # is done live in the application after migration (e.g. system field relations)
        return True

    def run(self, entries, cleanup=False):
        """Load entries."""
        for entry in entries:
            if self._validate(entry):
                self._prepare(entry)
                self._load(entry)

        if cleanup:
            self._cleanup()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration transform errors."""


class MultipleActionMatches(Exception):
    """More than one match for one transaction."""

    def __init__(self, tx):
        """Constructor."""
        super().__init__(f"Multiple action matches for {tx}")


class NoActionMatch(Exception):
    """No action match for a transaction."""

    def __init__(self, tx):
        """Constructor."""
        super().__init__(f"Could not detect action for {tx}")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration requests load module."""

from ...load.postgresql.bulk import PostgreSQLCopyLoad
from .table_generator import RequestTableGenerator


class RequestCopyLoad(PostgreSQLCopyLoad):
    """PostgreSQL COPY load."""

    def __init__(self, **kwargs):
        """Constructor."""
        super().__init__(table_generators=[RequestTableGenerator()], **kwargs)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Utils module."""

import json
from datetime import datetime
from uuid import UUID


def ts(iso=True, fmt=None):
    """Current timestamp string."""
    dt = datetime.now()
    if fmt:
        return dt.strftime(fmt)
    return dt.isoformat() if iso else dt.timestamp()


# PORT: from invenio_records.dictutils to avoid having an invenio constraint
# it could cause troubles with sqlalchemy and psycompg version
def parse_lookup_key(lookup_key):
    """Parse a lookup key."""
    if not lookup_key:
        raise KeyError("No lookup key specified")

    # Parse the list of keys
    if isinstance(lookup_key, str):
        keys = lookup_key.split(".")
    elif isinstance(lookup_key, list):
        keys = lookup_key
    else:
        raise TypeError("lookup must be string or list")

    return keys


def dict_set(source, key, value):
    """Set a value into a dict via a dot-notated key."""
    keys = parse_lookup_key(key)
    parent = source
    for key in keys[:-1]:
        parent = parent.setdefault(key, {})
    parent[keys[-1]] = value



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Requests transform tests."""

import dictdiffer

from invenio_rdm_migrator.streams.requests import (
    InclusionRequestEntry,
    RequestEntry,
    RequestTransform,
)


class MockRequestTransform(RequestTransform):
    """Mock class for request data transformation."""

    def _request(self, entry):
        """Transform the request."""
        return {"key": "value"}


def test_request_transform():
    """Test the full request transformation."""
    result = MockRequestTransform()._transform({})
    expected = {"key": "value"}
    assert not list(dictdiffer.diff(result, expected))


class MockRequestEntry(RequestEntry):
    """Transform a single mock request entry."""

    def _created(self, entry):
        """Returns the creation date of the request."""
        return "01/01/2023"

    def _updated(self, entry):
        """Returns the update date of the request."""
        return "01/01/2023"

    def _version_id(self, entry):
        """mock version."""
        return "1"

    def _json(self, entry):
        """Returns the request json body."""
        return {}

    def _number(self, entry):
        """Returns the request number."""
        return "1"

    def _expires_at(self, entry):
        """Returns the request expiration date."""
        return "01/01/2023"


def test_request_entry():
    """Test the request entry transformation."""
    result = MockRequestEntry().transform({})
    expected = {
        "created": "01/01/2023",
        "updated": "01/01/2023",
        "version_id": "1",
        "json": {},
        "number": "1",
        "expires_at": "01/01/2023",
    }

    assert not list(dictdiffer.diff(result, expected))


class MockInclusionRequestEntry(InclusionRequestEntry, MockRequestEntry):
    """Transform a single mock request entry."""

    # JSON related functions
    def _title(self, entry):
        """Returns the request title."""
        return "title"

    def _topic(self, entry):
        """Returns the request topic."""
        return "topic"

    def _receiver(self, entry):
        """Returns the request receiver."""
        return "receiver"

    def _created_by(self, entry):
        """Returns the request creation reference."""
        return "someone"


def test_inclusion_request_entry():
    """Test the request entry transformation."""
    result = MockInclusionRequestEntry().transform({})
    expected = {
        "created": "01/01/2023",
        "updated": "01/01/2023",
        "version_id": "1",
        "json": {
            "type": "community-inclusion",
            "title": "title",
            "topic": "topic",
            "status": "submitted",
            "receiver": "receiver",
            "created_by": "someone",
            "$schema": "local://requests/request-v1.0.0.json",
        },
        "number": "1",
        "expires_at": "01/01/2023",
    }

    assert not list(dictdiffer.diff(result, expected))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Communities models."""

from dataclasses import InitVar
from uuid import UUID

from sqlalchemy.orm import Mapped, mapped_column

from ...load.postgresql.models import Model


class Community(Model):
    """Community dataclass model."""

    __tablename__: InitVar[str] = "communities_metadata"

    id: Mapped[UUID] = mapped_column(primary_key=True)
    created: Mapped[str]  # datetime
    updated: Mapped[str]  # datetime
    json: Mapped[dict] = mapped_column(nullable=True)
    version_id: Mapped[int]
    slug: Mapped[str]
    bucket_id: Mapped[UUID] = mapped_column(primary_key=True)


class RDMParentCommunityMetadata(Model):
    """RDM Community Parent Metadata dataclass model."""

    __tablename__: InitVar[str] = "rdm_parents_community"

    community_id: Mapped[UUID] = mapped_column(primary_key=True)
    record_id: Mapped[UUID] = mapped_column(primary_key=True)
    request_id: Mapped[UUID]


class CommunityMember(Model):
    """Community members dataclass model."""

    __tablename__: InitVar[str] = "communities_members"

    id: Mapped[UUID] = mapped_column(primary_key=True)
    created: Mapped[str]  # datetime
    updated: Mapped[str]  # datetime
    json: Mapped[dict] = mapped_column(nullable=True)
    version_id: Mapped[int]
    role: Mapped[str]
    visible: Mapped[bool]
    active: Mapped[bool]
    community_id: Mapped[UUID]
    user_id: Mapped[int]
    group_id: Mapped[int] = mapped_column(nullable=True)
    request_id: Mapped[UUID] = mapped_column(nullable=True)


class FeaturedCommunity(Model):
    """Featured community dataclass model."""

    __tablename__: InitVar[str] = "communities_featured"

    community_id: Mapped[UUID]
    id: Mapped[int] = mapped_column(primary_key=True)
    created: Mapped[str]  # datetime
    updated: Mapped[str]  # datetime
    start_date: Mapped[str]  # datetime


class CommunityFile(Model):
    """Community file dataclass model."""

    __tablename__: InitVar[str] = "communities_files"

    created: Mapped[str]  # datetime
    updated: Mapped[str]  # datetime
    id: Mapped[UUID] = mapped_column(primary_key=True)
    json: Mapped[dict] = mapped_column(nullable=True)
    version_id: Mapped[int]
    key: Mapped[str] = mapped_column(primary_key=True)
    record_id: Mapped[UUID]
    object_version_id: Mapped[UUID]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
#
# Invenio-RDM-Migrator is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio RDM migration extract interfaces."""

from .base import Extract


class NullExtract(Extract):
    """Extract class to not read input data."""

    def run(self):
        """Return None only once.

        Subsequent calls to run will raise StopIteration.
        """
        # yield a None value
        yield
